# Time-stamp: <2021-05-10 19:10 ycopin@lyonovae03>

"""
spectrogrism
============

Available submodules
--------------------

* :mod:`spectrogrism.spectrogrism` - core functionalities
* :mod:`spectrogrism.distortion` - dirtortions
* :mod:`spectrogrism.snifs` - SNIFS specificities
* :mod:`spectrogrism.nisp` - NISP specificities
"""

__version__ = '0.9'

# What does "from spectrogrism import *" does?
__all__ = ['spectrogrism', 'distortion', 'snifs', 'nisp']

# Set to True if you want to import all previous modules directly
__importAll = True

if __importAll:
    for pkg in __all__:
        __import__(__name__ + '.' + pkg, fromlist=[None])

#



# -*- coding: utf-8 -*-
# Time-stamp: <2021-05-11 18:01:36 ycopin>

"""
spectrogrism
------------

Generic utilities for modeling grism-based spectrograph.

.. autosummary::

   Configuration
   OptConfig
   SimConfig
   SourceXDA
   DetectorPositionsXDA
   DetectorPositions
   Material
   Lens
   CameraOrCollimator
   Collimator
   Camera
   Telescope
   Prism
   Grating
   Grism
   Detector
   Spectrograph

.. inheritance-diagram:: spectrogrism.spectrogrism
"""

import warnings
from collections import OrderedDict
from functools import partial
import cmath as C

import numpy as N
import xarray as X

import spectrogrism.distortion as D

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>"

# Print options
LINEWIDTH = 70
N.set_printoptions(linewidth=LINEWIDTH, threshold=10)
X.set_options(display_width=LINEWIDTH)

# Constants ===============================================

RAD2DEG = 57.29577951308232     #: Convert from radians to degrees
RAD2MIN = RAD2DEG * 60          #: Convert from radians to arc minutes
RAD2SEC = RAD2MIN * 60          #: Convert from radians to arc seconds


# Technical Classes ==========================================

class Configuration(OrderedDict):
    """
    A :class:`OrderedDict`-derived configuration.

    .. autosummary::

       override
       save
       load
    """

    conftype = 'Configuration'            #: Configuration type

    def __init__(self, adict):
        """Initialize from dictionary `adict`."""

        OrderedDict.__init__(self, adict)
        self.name = self.get('name', 'default')

    def __str__(self):

        s = [f" {self.conftype} {self.name!r} ".center(LINEWIDTH, '-')]
        s += [ f"  {str(key):20}: {self[key]}" for key in list(self.keys()) ]

        return '\n'.join(s)

    def override(self, adict):
        """Override configuration from dictionary `adict`."""

        # warnings.warn(f"Overriding configuration {self.name!r} "
        #               f"with test values {adict}")

        self.name = adict.pop('name',
                              self.name + ' (overrided)'
                              if not self.name.endswith(' (overrided)')
                              else self.name)
        self.update(adict)

    def save(self, yamlname):
        """Save configuration to YAML file `yamlname`."""

        import yaml

        with open(yamlname, 'w') as yamlfile:
            yaml.dump(self, yamlfile)

        print(f"Configuration {self.name!r} saved in {yamlname!r}")

    @classmethod
    def load(cls, yamlname):
        """Load configuration from YAML file `yamlname`."""

        import yaml

        with open(yamlname, 'r') as yamlfile:
            self = yaml.load(yamlfile)

        print(f"Configuration {self.name!r} loaded from {yamlname!r}")

        return self

    def _repr_html_(self):
        """Pretty-printing in ipython notebooks."""

        html = ["<table>"]
        html.append(f"<caption>{self.conftype} {self.name}</caption>")
        for key in list(self.keys()):
            html.append(
                f"<tr><td><pre>{key}</pre></td><td>{self[key]}</td></tr>")
        html.append("</table>")
        return ''.join(html)

    def read_coeffs(self, name, start=0):
        """Read keys 'name{i}' and return list `[name0, name1, ...]`."""

        try:
            # List all keys starting with 'name', and get maximum index
            imax = max( int(key[len(name):])
                        for key in self if key.startswith(name) )
        except ValueError:      # No such key: return empty list
            # print(f"No {name!r} parameters")
            return []
        else:                   # Construct list from start to imax
            print(f"Reading {name + '#i'!r} parameters for i={start}...{imax}")
            return [ self.get(name + str(i), 0)
                     for i in range(start, imax + 1) ]


class OptConfig(Configuration):
    """
    Optical configuration.
    """

    conftype = "Optical configuration"

    @property
    def wref(self):
        """Central wavelength."""

        return self.get('wave_ref', 0.5 * sum(self['wave_range']))


class SimConfig(Configuration):
    """
    Simulation configuration.

    .. autosummary::

       get_wavelengths
       get_coordinates
    """

    conftype = "Simulation configuration"

    def get_wavelengths(self, optcfg):
        """Simulated wavelengths."""

        if 'WAVES' in self:
            waves = N.atleast_1d(self['WAVES'])
        else:
            npx = self.get('wave_npx', 1)
            if npx > 1:
                wmin, wmax = self.get('wave_range', optcfg['wave_range'])
                waves = N.linspace(wmin, wmax, npx)
            elif npx == 1:
                waves = optcfg.wref
            else:
                raise ValueError(f"Invalid number of pixels: {npx}")

        return waves

    def get_coordinates(self):
        """Simulated input complex coordinates `[ x + 1j*y ]`."""

        incoords = N.atleast_1d(self.get('input_coords', 0))
        if N.ndim(incoords) == 1:
            # [x]: generate square sampling x × x
            x, y = N.meshgrid(incoords, incoords)
            coords = (x + 1j * y).ravel()
        elif N.ndim(incoords) == 2 and N.shape(incoords)[1] == 2:
            # [[x, y]]: arbitrary sampling
            coords = incoords[:, 0] + 1j * incoords[:, 1]
        else:
            raise NotImplementedError("Unsupported input coordinates.")

        # Apply rotation if any
        angle = self.get('input_angle', 0)   # [rad]
        if angle:
            coords *= C.exp(1j * angle)

        return coords


def SourceXDA(wavelengths, signal=1, zcoord=0j, name=None):
    """
    A 1D :class:`xarray.DataArray`.

    * values: signal (float)
    * indices: `wavelength` (float)
    * attrs: `zcoord` (complex)

    :param list wavelengths: strictly increasing input wavelengths [m]
    :param list signal: input signal [arbitrary units]
    :param str name: optional spectrum name (e.g. "Grism transmission")
    """

    assert (N.diff(wavelengths) > 0).all(), \
        "Wavelengths not strictly increasing."

    return X.DataArray(signal,
                       coords=(('wavelength', wavelengths),),
                       attrs=dict(zcoord=zcoord),
                       name=name)


def str_SourceXDA(self):
    """
    Return a SourceXDA string.
    """

    w = self.coords['wavelength'].data    # N.array
    z = self.attrs['zcoord']

    return (f"{self.name + ': ' if self.name else ''}{len(self):d} "
            "wavelengths in "
            f"[{w[0]*1e6:.2f}, {w[-1]*1e6:.2f}] µm "
            f"at ({z.real:+.2f}, {z.imag:+.2f}) mm")


def DetectorPositionsXDA(wavelengths, coordinates, zout=0):
    """
    A 2D :class:`xarray.DataArray`.

    * values: output coordinates (complex)
    * indices: `wavelength` (float) and `coordinates` (complex)

    :param list wavelengths: strictly increasing input wavelengths [m]
    :param list coordinates: input complex coordinates [m]
    :param list zout: output (wavelengths, zin) complex coordinates [m]
    """

    assert (N.diff(wavelengths) > 0).all(), \
        "Wavelengths not strictly increasing."

    return X.DataArray(zout, coords=(('wavelength', wavelengths),
                                     ('coordinate', coordinates),))


class DetectorPositions:
    """
    **DEPRECATED DOCSTRING**

    A container for complex 2D-positions on the detector.

    * items are observational modes (dispersion orders or photometric bands)
    * major axis is wavelength (shared among all modes)
    * minor axis is input complex coordinates (shared among all modes)

    For each mode, the corresponding :class:`pandas.DataFrame` is therefore
    organized the following way::

                zin_1     zin_2  ...     zin_N
      lbda1  zout_1_1  zout_2_1  ...  zout_N_1
      lbda2  zout_1_2  zout_2_2  ...  zout_N_2
      ...
      lbdaM  zout_1_M  zout_2_M  ...  zout_N_M

    * `self.modes` returns the available observing modes;
    * `df = self.panel[mode]` is the dataframe corresponding to a given
      observing `mode`;
    * `self.wavelengths` returns wavelength array `[lbdas]`;
    * `self.coordinates` returns (complex) input coordinate array `[zins]`;
    * `df[zin]` returns a wavelength-indexed :class:`pandas.Series` of complex
      detector positions.

    .. WARNING:: Indexing by float is not precisely a good idea... Float
       indices (wavelengths and input coordinates) are therefore rounded first
       with a sufficient precision (e.g. nm for wavelengths).

    .. TODO:: do not use complex coordinates as column name (this is not
       natively supported in Pandas, nor a good idea). Store input coordinates
       in an associated :class:`pandas.Series`, and use series index as column
       names.

    .. autosummary::

       add_mode
       plot
       check_alignment
       compute_offset
       compute_rms
    """

    mdict = {-1: 'D', 0: '.', 1: 'o', 2: 's',
             'Y': 'D', 'J': 'o', 'H': 's'}

    def __init__(self, wavelengths=None, coordinates=None,
                 spectrograph=None, name='default'):
        """
        Initialize *empty* container from spectrograph and wavelength array.

        :param list wavelengths: input wavelengths [m]
        :param list coordinates: input complex coordinates [m]
        :param Spectrograph spectrograph: associated spectrograph (if any)
        :param str name: informative label
        """

        if wavelengths is None:
            wavelengths = [1e-6]
        else:
            assert (N.diff(wavelengths) > 0).all(), \
                "Wavelengths not strictly increasing."

        if coordinates is None:
            coordinates = [0j]

        self.xda = X.DataArray(0,
                               coords=(('mode', []),
                                       ('wavelength', wavelengths),
                                       ('coordinate', N.array(coordinates,
                                                              dtype=complex))))

        if spectrograph is not None:
            assert isinstance(spectrograph, Spectrograph), \
                "spectrograph should be a Spectrograph."
        self.spectrograph = spectrograph

        self.name = name

    @property
    def modes(self):
        """Observational mode array."""

        return self.xda.coords['mode'].data

    @property
    def coordinates(self):
        """Input source (complex) coordinate array."""

        return self.xda.coords['coordinate'].data.astype(complex)

    @property
    def wavelengths(self):
        """Wavelength array."""

        return self.xda.coords['wavelength'].data

    def add_mode(self, mode, xda):
        """Add *xda* as observational *mode* to current panel."""

        assert mode in self.mdict, f"Unknown mode {mode!r}."

        self.xda = X.concat((self.xda,
                             xda.expand_dims(mode=[mode], axis=0)),
                            dim='mode', compat='equals')

    def plot(self, ax=None, coordinates=None, modes=None, blaze=False,
             subsampling=1, **kwargs):
        """
        Plot spectra on detector plane.

        :param matplotlib.pyplot.Axes ax: pre-existing axes instance if any
        :param list coordinates: selection of input coordinates to be plotted
        :param list modes: selection of observing modes to be plotted
        :param bool blaze: encode the blaze function in the marker size
        :param int subsampling: sub-sample coordinates and wavelengths
        :param kwargs: options propagated to
            :func:`matplotlib.pyplot.Axes.scatter`
        :return: :class:`matplotlib.pyplot.Axes`
        """

        import matplotlib.pyplot as P

        if ax is None:
            fig = P.figure()
            ax = fig.add_subplot(1, 1, 1,
                                 xlabel="x [mm]", ylabel="y [mm]",
                                 title=self.name)
            ax.set_aspect('equal', adjustable='datalim')
        else:
            fig = None          # Will serve as a flag

        kwargs.setdefault('edgecolor', 'none')
        kwargs.setdefault('label', self.name)

        # Input observational modes
        if modes is None:             # Plot all observational modes
            modes = self.modes

        # Input coordinates
        if coordinates is None:       # Plot all spectra
            xys = self.coordinates
        else:
            xys = coordinates         # Plot specific spectra

        if subsampling > 1:
            xys = xys[::subsampling]

        # Input wavelengths
        if subsampling > 1:
            lbda = self.wavelengths[::subsampling]
        else:
            lbda = self.wavelengths

        bztrans = N.ones_like(lbda)     # Default blaze transmission
        # Color has to have same shape as coordinates, (nlbda, ncoords)
        c = N.broadcast_to(lbda[:, N.newaxis], (len(lbda), len(xys)))

        # Segmented colormap (blue to red)
        cmap = P.matplotlib.colors.LinearSegmentedColormap(
            'dummy', P.get_cmap('Spectral_r')._segmentdata,
            len(self.wavelengths))

        for mode in modes:              # Loop over observational modes
            try:
                xda = self.xda.sel(mode=mode)
            except KeyError:
                warnings.warn(
                    f"{str_mode(mode)} not in {self.name!r}, skipped")
                continue

            kwcopy = kwargs.copy()
            marker = kwcopy.pop('marker', self.mdict[mode])
            kwcopy['label'] += f" {'#' if is_spectred(mode) else ''}{mode}"

            if blaze and self.spectrograph:
                bztrans = self.spectrograph.grism.blaze_function(lbda, mode)
                s = kwcopy.pop('s', N.maximum(60 * N.sqrt(bztrans), 10))
                # Size has to have same shape as coordinates, (nlbda, ncoords)
                s = N.broadcast_to(s[:, N.newaxis], (len(lbda), len(xys)))
            else:
                s = kwcopy.pop('s', 40)

            # Raise a KeyError if one of the coordinates is unknown
            positions = xda.sel(coordinate=xys).values * 1e3  # Positions [mm]
            if subsampling > 1:
                positions = positions[::subsampling]

            sc = ax.scatter(positions.real, positions.imag,
                            c=c * 1e6,   # Wavelength [µm]
                            cmap=cmap, s=s, marker=marker, **kwcopy)

            # kwargs.pop('label', None)      # Label only for one mode

        if fig:                 # Newly created axes
            fig.colorbar(sc, label="Wavelength [µm]")

        return ax

    def check_alignment(self, other):
        """
        Check compatibility in wavelengths and positions with other instance.

        :param DetectorPositions other: other instance to be compared to
        :raise IndexError: incompatible instance
        """

        if not N.allclose(other.wavelengths, self.wavelengths):
            raise IndexError(f"{self.name!r} and {other.name!r} "
                             "have incompatible wavelengths.")

        if not N.allclose(other.coordinates, self.coordinates):
            raise IndexError(f"{self.name!r} and {other.name!r} "
                             "have incompatible coordinates.")

    def compute_offset(self, other, mode=1):
        """
        Compute (complex) position offsets to other instance.

        :param DetectorPositions other: other instance to be compared to
        :param mode: observing mode (dispersion order or photometric band)
        :return: (complex) position offsets [m]
        :rtype: :class:`pandas.DataFrame`
        :raise KeyError: requested mode cannot be found

        .. Warning:: `self` and `other` are supposed to be compatible
           (see :func:`test_compatibility`).
        """

        # Dataframe of (complex) position offsets for requested mode
        return other.xda.sel(mode=mode) - self.xda.sel(mode=mode)

    def compute_rms(self, other, mode=1):
        """
        Compute total RMS distance to other instance.

        :param DetectorPositions other: other instance to be tested
        :param mode: observing mode (dispersion order or photometric band)
        :return: RMS offset [m]
        :rtype: float
        :raise KeyError: requested mode cannot be found

        .. Warning:: `self` and `other` are supposed to be compatible
           (see :func:`test_compatibility`).
        """

        offsets = N.abs(self.compute_offset(other, mode=mode))

        return (offsets.values ** 2).mean(axis=None) ** 0.5


class Material:
    """
    Optical material.

    The refractive index is described by its Sellmeier coefficients.

    **Reference:** `Sellmeier equation
    <https://en.wikipedia.org/wiki/Sellmeier_equation>`_

    .. autosummary::

       index
    """

    #: Sellmeier coefficients [B1, B2, B3, C1, C2, C3] of known materials.
    materials = dict(
        # Glasses
        BK7=[ 1.03961212,    2.31792344e-1, 1.01046945,
              6.00069867e-3, 2.00179144e-2, 1.03560653e+2],
        UBK7=[1.01237433,    2.58985218e-1, 1.00021628,
              5.88328615e-3, 1.90239921e-2, 1.04079777e+2],
        SF4=[ 1.61957826,    3.39493189e-1, 1.02566931,
              1.25502104e-2, 5.33559822e-2, 1.1765222e+2],
        SK5=[ 0.99146382,    4.95982121e-1, 0.98739392,
              5.22730467e-3, 1.72733646e-2, 0.983594579e+2],
        F2=[  1.34533359,    2.09073176e-1, 0.93735716,
              9.97743871e-3, 4.70450767e-2, 1.11886764e+2],
        SF57=[1.81651371,    4.28893641e-1, 1.07186278,
              1.43704198e-2, 5.92801172e-2, 1.21419942e+2],
        # Fused silica
        FS=[  0.6961663,     4.079426e-1,   0.8974794,
              4.679148e-3,   1.351206e-2,  97.9340],
        # Epoxy
        EPR=[ 0.512479,      0.838483,     -0.388459,
             -0.0112765,     0.0263791,   557.682],
        EPB=[ 0.406836,      1.03517,      -0.140328,
             -0.0247382,     0.0261501,   798.366],
        # Null material (refractive index = 1)
        null=[0, 0, 0, 0, 0, 0],
    )

    def __init__(self, name):
        """
        Initialize material from its name.

        :param str name: material name (should be in :attr:`Material.materials`)
        :raise KeyError: unknown material name
        """

        try:
            #: Sellmeier coefficients `[B1, B2, B3, C1, C2, C3]`
            self.coeffs = self.materials[name]
        except KeyError as error:
            raise KeyError(f"Unknown material {name!r}") from error
        self.name = name        #: Name of the material

    def __str__(self):

        return f"Material: {self.name}, n(1 µm)={self.index(1e-06):.3f}"

    def index(self, wavelengths):
        r"""
        Compute refractive index from Sellmeier expansion.

        Sellmeier expansion for refractive index:

        .. math:: n(\lambda)^2 = 1 + \sum_{i}\frac{B_i\lambda^2}{\lambda^2-C_i}

        with :math:`\lambda` in microns.

        :param numpy.ndarray wavelengths: wavelengths [m]
        :return: refractive index
        """

        if self.name == 'null':
            return N.ones_like(wavelengths)

        lmu2 = (wavelengths * 1e6) ** 2        # (wavelength [µm])**2
        n2m1 = N.sum([ b / (1 - c / lmu2)       # n**2 - 1
                       for b, c in zip(self.coeffs[:3], self.coeffs[3:]) ],
                     axis=0)

        return N.sqrt(n2m1 + 1)


# Optical element classes =================================

class Lens:
    """
    Convert a 2D-position into a 2D-position.

    .. autosummary:

       forward
       backward
    """

    def __init__(self, gamma, gdist=None, cdist=None):
        """
        Initialize the element from its optical parameters.

        :param float gamma: transverse 'grandissement'
        :param D.GeometricDistortion gdist: geometric distortion
        :param D.ChromaticDistortion cdist: chromatic distortion (lateral color)
        """

        self.gamma = float(gamma)              #: Transverse 'grandissement'

        if gdist is not None:
            assert isinstance(gdist, D.GeometricDistortion), \
                "gdist should be a GeometricDistortion."
        else:
            gdist = D.GeometricDistortion()        # Null geometric distortion

        if cdist is not None:
            assert isinstance(cdist, D.ChromaticDistortion), \
                "cdist should be a ChromaticDistortion."
        else:
            cdist = D.ChromaticDistortion()        # Null lateral color

        self.gdist = gdist              #: Geometric distortion
        self.cdist = cdist              #: Chromatic distortion (lateral color)

    def forward(self, positions):

        raise NotImplementedError()

    def backward(self, positions):

        raise NotImplementedError()


class CameraOrCollimator:
    """
    An optical element converting to and fro directions and positions.
    """

    def __init__(self, flength, gdist=None, cdist=None):
        """
        Initialize the element from its optical parameters.

        :param float flength: focal length [m]
        :param D.GeometricDistortion gdist: geometric distortion
        :param D.ChromaticDistortion cdist: chromatic distortion (lateral color)
        """

        self.flength = float(flength)              #: Focal length [m]

        if gdist is not None:
            assert isinstance(gdist, D.GeometricDistortion), \
                "gdist should be a GeometricDistortion."
        else:
            gdist = D.GeometricDistortion()        # Null geometric distortion

        if cdist is not None:
            assert isinstance(cdist, D.ChromaticDistortion), \
                "cdist should be a ChromaticDistortion."
        else:
            cdist = D.ChromaticDistortion()        # Null lateral color

        self.gdist = gdist              #: Geometric distortion
        self.cdist = cdist              #: Chromatic distortion (lateral color)

    def __str__(self):

        s = f"{self.__class__.__name__}: f={self.flength:.3f} m"
        s += f'\n  {self.gdist}'
        s += f'\n  {self.cdist}'

        return s


class Collimator(CameraOrCollimator):
    """
    Convert a 2D-position (in the focal plane) into a 2D-direction.

    .. autosummary::

       forward
       backward
    """

    def __init__(self, config):
        """
        Initialize from optical configuration.

        :param OptConfig config: optical configuration
        :raise KeyError: missing configuration key
        """

        try:
            flength = config['collimator_flength']
            center = (config.get('collimator_gdist_x0', 0) +
                      config.get('collimator_gdist_y0', 0) * 1j )
            Kcoeffs = config.read_coeffs('collimator_gdist_K', start=1)
            Pcoeffs = config.read_coeffs('collimator_gdist_P', start=1)
            lcoeffs = config.read_coeffs('collimator_cdist_C', start=1)
        except KeyError as err:
            raise KeyError("Invalid configuration file: "
                           f"missing key {err.args[0]!r}") from err
        else:
            gdist = D.GeometricDistortion(
                center, Kcoeffs=Kcoeffs, Pcoeffs=Pcoeffs)
            cdist = D.ChromaticDistortion(config.wref, lcoeffs)

        super().__init__(flength, gdist, cdist)  # CameraOrCollimator

    def forward(self, positions, wavelengths, gamma):
        r"""
        Forward light propagation through the collimator.

        .. math:

           z' = D(-z / f) - z/f \times L(\lambda) / \gamma

        where:

        * :math:`z = x + iy` is the complex position
        * *D* is the :class:`D.GeometricDistortion` geometric distortion
        * *L* is the :class:`D.ChromaticDistortion` lateral color
        * :math:`\gamma` is the spectrograph magnification (fcam/fcoll)

        The geometric distortion is therefore applied on the (flipped)
        positions.
        """

        zu = - positions / self.flength  # Undistorted (flipped) directions
        zd = self.gdist.forward(zu)      # Distorted directions
        if self.cdist:                   # Add lateral colors
            zd += zu * self.cdist.amplitude(wavelengths) / gamma

        return zd

    def backward(self, directions, wavelength, gamma):
        """
        Backward light propagation through the collimator.

        See :func:`Collimator.forward` for details.
        """

        if self.cdist:                   # Lateral colors
            lcol = self.cdist.amplitude(wavelength) / gamma  # scalar
        else:
            lcol = 0.

        return - self.flength * self.gdist.backward(directions, lcol=lcol)


class Camera(CameraOrCollimator):
    """
    Convert a 2D-direction into a 2D-position (in the detector plane).

    .. Note:: the detector coordinate axes are flipped, so that
       sources remain in the same quadrant in the focal *and* detector
       planes.

    .. autosummary::

       forward
       backward
    """

    def __init__(self, config):
        """
        Initialize from optical configuration.

        :param OptConfig config: optical configuration
        :raise KeyError: missing configuration key
        """

        try:
            flength = config['camera_flength']
            center = (config.get('camera_gdist_x0', 0) +
                      config.get('camera_gdist_y0', 0) * 1j )
            Kcoeffs = config.read_coeffs('camera_gdist_K', start=1)
            Pcoeffs = config.read_coeffs('camera_gdist_P', start=1)
            lcoeffs = config.read_coeffs('camera_cdist_C', start=1)
        except KeyError as err:
            raise KeyError("Invalid configuration file: "
                           f"missing key {err.args[0]!r}") from err
        else:
            gdist = D.GeometricDistortion(
                center, Kcoeffs=Kcoeffs, Pcoeffs=Pcoeffs)
            cdist = D.ChromaticDistortion(config.wref, lcoeffs)

        super().__init__(flength, gdist, cdist)  # CameraOrCollimator

    def forward(self, directions, wavelengths):
        r"""
        Forward light propagation through the camera.

        .. math:

           z' = -[D(z) + z L(\lambda)] \times f

        where:

        * :math:`z = x + iy` is the complex position
        * *D* is the :class:`D.GeometricDistortion` geometric distortion
        * *L* is the :class:`D.ChromaticDistortion` lateral color

        The geometric distortion is therefore applied on the (input) directions.
        """

        zd = self.gdist.forward(directions)  # Distorted (unflipped) directions
        if self.cdist:
            zd += directions * self.cdist.amplitude(wavelengths)

        return - zd * self.flength           # Distorted (flipped) positions

    def backward(self, positions, wavelength):
        """
        Backward light propagation through the camera.

        See :func:`Camera.forward` for details.
        """

        if self.cdist:                   # Lateral colors
            lcol = self.cdist.amplitude(wavelength)  # scalar
        else:
            lcol = 0.

        return self.gdist.backward(- positions / self.flength, lcol=lcol)


class Telescope(Camera):
    """
    All-reflective telescope (no chromatic distortions).

    Convert a 2D-direction in the sky into a 2D-position in the focal plane.
    """

    def __init__(self, config):
        """
        Initialize from optical configuration.

        :param OptConfig config: optical configuration
        :raise KeyError: missing configuration key
        """

        try:
            flength = config['telescope_flength']
            center = (config.get('telescope_gdist_x0', 0) +
                      config.get('telescope_gdist_y0', 0) * 1j )
            Kcoeffs = config.read_coeffs('telescope_gdist_K', start=1)
            Pcoeffs = config.read_coeffs('telescope_gdist_P', start=1)
        except KeyError as err:
            raise KeyError("Invalid configuration file: "
                           f"missing key {err.args[0]!r}") from err
        else:
            gdist = D.GeometricDistortion(
                center, Kcoeffs=Kcoeffs, Pcoeffs=Pcoeffs)

        # Initialize from CameraOrCollimator parent class (without
        # chromatic aberration)
        super(Camera, self).__init__(flength, gdist, cdist=None)  # CameraOrColl

        # Remove explicit dependency to wavelengths in Camera.forward
        # and backward
        self.forward = partial(super().forward, wavelengths=None)
        self.backward = partial(super().backward, wavelength=None)


class Prism:
    """
    A triangular transmissive prism.

    .. Note::

       - The entry surface is roughly perpendicular (up to the tilt
         angles) to the optical axis Oz.
       - The apex (prism angle) is aligned with the *x*-axis

    .. autosummary::

       rotation
       rotation_x
       rotation_y
       rotation_z
       refraction
    """

    def __init__(self, angle, material, tilts=[0, 0, 0]):
        """
        Initialize grism from its optical parameters.

        :param float angle: prism angle [rad]
        :param Material material: prism material
        :param 3-list tilts: prism tilts (x, y, z) [rad]
        """

        assert isinstance(material, Material), "material should be a Material."
        assert len(tilts) == 3, "tilts should be a length-3 list."

        self.angle = angle                #: Prism angle [rad]
        self.material = material          #: Prism material
        self.tilts = tilts                #: Prism tilts (x, y, z) [rad]

    def __str__(self):

        # Present tilt angles in arcmin
        tilts = ','.join( f"{t * RAD2MIN:+.0f}'" for t in self.tilts )

        return (f"Prism [{self.material.name}]: "
                f"A={self.angle * RAD2DEG:.2f}°, tilts={tilts}")

    @property
    def tiltx(self):
        """
        Expose prism x-tilt [rad], rotation around the prism apex/grating
        grooves.
        """

        return self.tilts[0]

    @tiltx.setter
    def tiltx(self, xtilt):

        self.tilts[0] = xtilt

    @property
    def tilty(self):
        """
        Expose prism y-tilt [rad].
        """

        return self.tilts[1]

    @tilty.setter
    def tilty(self, ytilt):

        self.tilts[1] = ytilt

    @property
    def tiltz(self):
        """
        Expose prism z-tilt [rad], rotation around the optical axis.
        """

        return self.tilts[2]

    @tiltz.setter
    def tiltz(self, ztilt):

        self.tilts[2] = ztilt

    @staticmethod
    def rotation(x, y, theta):
        """
        2D-rotation of position around origin with direct angle `theta` [rad].
        """

        # Rotation in the complex plane
        p = (N.array(x) + 1j * N.array(y)) * C.exp(1j * theta)

        return (p.real, p.imag)

    @classmethod
    def rotation_x(cls, xyz, theta):
        """Rotation around x-axis."""

        x, y, z = xyz
        y, z = cls.rotation(y, z, theta)
        return N.vstack((x, y, z)).squeeze()

    @classmethod
    def rotation_y(cls, xyz, theta):
        """Rotation around y-axis."""

        x, y, z = xyz
        x, z = cls.rotation(x, z, theta)
        return N.vstack((x, y, z)).squeeze()

    @classmethod
    def rotation_z(cls, xyz, theta):
        """Rotation around z-axis."""

        x, y, z = xyz
        x, y = cls.rotation(x, y, theta)
        return N.vstack((x, y, z)).squeeze()

    @staticmethod
    def refraction(xyz, n1, n2):
        """
        Refraction law from medium 1 to medium 2, by plane interface
        :math:`(Oxy)`.

        :param 3-tuple xyz: input 3D-direction (from medium 1)
        :param float n1: Refractive index of medium 1
        :param float n2: Refractive index of medium 2
        :return: output 3D-direction (to medium 2)
        :rtype: 3-tuple
        """

        x1, y1, _ = xyz
        x2 = x1 * n1 / n2
        y2 = y1 * n1 / n2
        z2 = N.sqrt(1 - (x2 ** 2 + y2 ** 2))

        return N.vstack((x2, y2, z2)).squeeze()


class Grating:
    """
    A transmissive grating.

    The grooves of the (transmission) grating are aligned along *x*.

    .. autosummary::

       forward
       backward
    """

    def __init__(self, rho, material, blaze=0):
        """
        Initialize grating from its optical parameters.

        :param float rho: grating groove density [lines/mm]
        :param Material material: grating material
        :param float blaze: grating blaze angle [rad]
        """

        assert isinstance(material, Material), "material should be a Material."

        self.rho = rho            #: Grating groove density [lines/mm]
        self.material = material  #: Grating material
        self.blaze = blaze        #: Grating blaze angle [rad]

    def __str__(self):

        return (f"Grating [{self.material.name}]: "
                f"rho={self.rho:.1f} g/mm, blaze={self.blaze * RAD2DEG:.2f}°")

    def forward(self, xyz, wavelengths, order=1):
        """
        Forward light propagation through a grating.

        The propagation is done from material (*n*) to vacuum (*1*).

        :param 3-tuple xyz: input 3D-direction *(x, y, z)* [m]
        :param numpy.ndarray wavelengths: wavelengths [m]
        :param int order: dispersion order
        :return: output 3D-direction *(x', y', z')*
        :rtype: 3-tuple
        """

        n = self.material.index(wavelengths)  # Index of refraction

        x, y, _ = xyz
        xp = x * n
        yp = y * n + order * wavelengths * self.rho * 1e3
        zp = N.sqrt(1 - (xp ** 2 + yp ** 2))

        return N.vstack((xp, yp, zp)).squeeze()

    def backward(self, xyz, wavelength, order=1):
        """
        Backward light propagation through a grating.

        The propagation is done from vacuum (*1*) to material (*n*).

        :param 3-tuple xyz: output 3D-direction *(x', y', z')* [m]
        :param float wavelength: wavelength [m]
        :param int order: dispersion order
        :return: input 3D-direction *(x, y, z)*
        :rtype: 3-tuple
        """

        n = self.material.index(wavelength)  # Index of refraction

        xp, yp, _ = xyz
        x = xp / n
        y = (yp - order * wavelength * self.rho * 1e3) / n
        z = N.sqrt(1 - (x ** 2 + y ** 2))

        return N.vstack((x, y, z)).squeeze()


class Grism:
    """
    A :class:`Prism` and a :class:`Grating` on the exit surface.

    .. autosummary::

       blaze_function
       direction2xyz
       xyz2direction
       forward
       backward
       null_deviation
    """

    def __init__(self, config):
        """
        Initialize from optical configuration.

        :param OptConfig config: optical configuration
        :raise KeyError: missing configuration key
        """

        try:
            angle = config['grism_prism_angle']
            prism_material = config['grism_prism_material']
            tilts = [ config.get('grism_prism_tilt' + ax, 0) for ax in 'xyz' ]
            rho = config['grism_grating_rho']
            grating_material = config['grism_grating_material']
            blaze = config.get('grism_grating_blaze', 0)
        except KeyError as err:
            raise KeyError("Invalid configuration file: "
                           f"missing key {err.args[0]!r}") from err

        self.prism = Prism(angle, Material(prism_material), tilts=tilts)
        self.grating = Grating(rho, Material(grating_material), blaze)

    def __str__(self):

        return f"""Grism:
  {self.prism}
  {self.grating}
  1st-order null-deviation wavelength: {self.null_deviation(order=1)*1e6:.2f} µm
"""

    def blaze_function(self, wavelengths, order=1):
        r"""
        Blaze function.

        In the normal configuration, the blaze function of a grism is given by

        .. math:: B = \frac{\sin^2\Theta}{\Theta^2}

        with:

        - :math:`\rho \lambda \Theta = \pi \cos\gamma \times (n_g\sin
          i - \sin r)`, :math:`\gamma` being the blaze angle and
          :math:`\rho` the line density of the grating;
        - :math:`i = \alpha' - \gamma` where :math:`n_g \sin\alpha' =
          n_p\sin A` with :math:`n_p` and :math:`n_g` the refraction
          index of prism glass and grating resine respectively and
          :math:`A` the grism angle
        - :math:`r = \beta - \gamma` where :math:`\sin\beta = n_p\sin
          A - m\rho\lambda` with :math:`m` the diffraction order.

        :param numpy.ndarray wavelengths: wavelengths [m]
        :param int order: dispersion order
        :return: blaze function (i.e. transmission at input wavelengths)
        :rtype: :class:`numpy.ndarray`
        """

        np = self.prism.material.index(wavelengths)    # Prism index
        ng = self.grating.material.index(wavelengths)  # Grating index

        rholbda = self.grating.rho * 1e3 * wavelengths  # g/m * m = unitless
        npsinA = np * N.sin(self.prism.angle)

        i = N.arcsin(npsinA / ng) - self.grating.blaze               # [rad]
        r = N.arcsin(npsinA - order * rholbda) - self.grating.blaze  # [rad]

        theta = (N.pi / rholbda * N.cos(self.grating.blaze) *
                 (ng * N.sin(i) - N.sin(r)))
        bf = (N.sin(theta) / theta) ** 2

        return bf

    @staticmethod
    def direction2xyz(direction):
        """
        Convert a 2D-direction into a 3D-direction (a unit vector).

        :param complex direction: 2D-direction
        :return: 3D-direction
        :type: 3-tuple
        """

        tantheta, phi = rect2pol(direction)
        tan2 = tantheta ** 2
        costheta = N.sqrt(1 / (1 + tan2))
        sintheta = N.sqrt(tan2 / (1 + tan2))

        return N.vstack((N.cos(phi) * sintheta,
                         N.sin(phi) * sintheta,
                         costheta)).squeeze()

    @staticmethod
    def xyz2direction(xyz):
        """
        Convert a 3D-direction (a unit vector) into a 2D-direction.

        :param 3-tuple xyz: 3D-direction
        :return: 2D-direction
        :rtype: complex
        """

        x, y, z = xyz
        tantheta = N.hypot(x, y) / z
        phi = N.arctan2(y, x)

        return pol2rect(tantheta, phi)

    def forward(self, direction, wavelengths, order=1):
        """
        Forward propagation through a grism (prism + grating).

        :param complex direction: 2D-direction [rad]
        :param numpy.ndarray wavelengths: wavelengths [m]
        :param int order: dispersion order
        :return: 2D-directions [rad]
        :rtype: :class:`numpy.ndarray`
        """

        # Convert input 2D-direction into a 3D-direction
        xyz = self.direction2xyz(direction)

        # Get aligned with prism
        xyz = self.prism.rotation_x(xyz, self.prism.tilts[0])
        xyz = self.prism.rotation_y(xyz, self.prism.tilts[1])
        xyz = self.prism.rotation_z(xyz, self.prism.tilts[2])

        # Vacuum/prism interface
        xyz = self.prism.refraction(xyz, 1,
                                    self.prism.material.index(wavelengths))

        # Prism angle
        xyz = self.prism.rotation_x(xyz, self.prism.angle)

        # Prism/grating interface
        xyz = self.prism.refraction(xyz,
                                    self.prism.material.index(wavelengths),
                                    self.grating.material.index(wavelengths))

        # Grating dispersion
        xyz = self.grating.forward(xyz, wavelengths, order=order)

        # Back to original orientation
        xyz = self.prism.rotation_x(xyz, -self.prism.angle)
        xyz = self.prism.rotation_z(xyz, -self.prism.tilts[2])
        xyz = self.prism.rotation_y(xyz, -self.prism.tilts[1])
        xyz = self.prism.rotation_x(xyz, -self.prism.tilts[0])

        # Convert output 3D-directions into 2D-directions
        direction = self.xyz2direction(xyz)

        return direction

    def backward(self, direction, wavelength, order=1):
        """
        Backward propagation through a grism (prism + grating).

        See :func:`Grism.forward` for parameters.
        """

        # Convert output 2D-direction into a 3D-direction
        xyz = self.direction2xyz(direction)

        # Get aligned with A-tilted grating
        xyz = self.prism.rotation_x(xyz, self.prism.tilts[0])
        xyz = self.prism.rotation_y(xyz, self.prism.tilts[1])
        xyz = self.prism.rotation_z(xyz, self.prism.tilts[2])
        xyz = self.prism.rotation_x(xyz, self.prism.angle)

        # Grating
        xyz = self.grating.backward(xyz, wavelength, order=order)

        # Grating/prism interface
        xyz = self.prism.refraction(xyz,
                                    self.grating.material.index(wavelength),
                                    self.prism.material.index(wavelength))

        # Prism
        xyz = self.prism.rotation_x(xyz, -self.prism.angle)

        # Prism/vacuum interface
        xyz = self.prism.refraction(xyz,
                                    self.prism.material.index(wavelength), 1)

        # Back to original orientation
        xyz = self.prism.rotation_z(xyz, -self.prism.tilts[2])
        xyz = self.prism.rotation_y(xyz, -self.prism.tilts[1])
        xyz = self.prism.rotation_x(xyz, -self.prism.tilts[0])

        # Convert intput 3D-direction into 2D-direction
        direction = self.xyz2direction(xyz)

        return direction

    def null_deviation(self, order=1):
        r"""
        Null-deviation wavelength (approximated) [m].

        This is the solution to:

        .. math:: m \rho \lambda = (n(\lambda) - 1) \sin(A)

        where:

        - *A*: grism angle [rad]
        - :math:`\rho`: groove density [line/mm]
        - :math:`n(\lambda)`: prism refractive index
        - *m*: dispersion order

        :param int order: dispersion order
        :return: null deviation wavelength [m]
        :raise RuntimeError: if not converging
        """

        import scipy.optimize as SO

        k = N.sin(self.prism.angle) / (order * self.grating.rho * 1e3)

        def objf(l):

            return l - k * (self.prism.material.index(l) - 1)

        lbda0 = SO.newton(objf, 1e-6)  # Look around 1 µm

        return lbda0


class Detector:
    """
    A simple translated and rotated detector.

    .. autosummary::

       forward
       backward
    """

    def __init__(self, config):
        """
        Initialize from optical configuration.

        :param OptConfig config: optical configuration
        :raise KeyError: missing configuration key
        """

        self.dx = config.get('detector_dx', 0)        #: X-offset [m]
        self.dy = config.get('detector_dy', 0)        #: Y-offset [m]
        self.angle = config.get('detector_angle', 0)  #: Rotation [rad]
        self.pxsize = config['detector_pxsize']       #: Pixel size [m]

    @property
    def dxdy(self):
        """
        Expose complex offset `dx + 1j*dy` [m].
        """

        return self.dx + 1j * self.dy     # Faster than complex(dx, dy)

    @dxdy.setter
    def dxdy(self, dxdy):

        self.dx, self.dy = dxdy.real, dxdy.imag

    def __str__(self):

        return f"""\
  Detector: pxsize={self.pxsize*1e6:.0f} µm
  Offset=({self.dx*1e3:+.3f}, {self.dy*1e3:+.3f}) mm, angle={self.angle*RAD2DEG:.1f}°
"""

    def forward(self, positions):
        """Forward propagation to detector."""

        return (positions + self.dxdy) * C.exp(1j * self.angle)

    def backward(self, positions):
        """Backward propagation from detector."""

        return positions / C.exp(1j * self.angle) - self.dxdy


class Spectrograph:
    """
    A collimated spectrograph.

    A :class:`Collimator`, a :class:`Grism` in a collimated beam, a
    :class:`Camera` and a :class:`Detector`, plus an optional
    :class:`Telescope`.

    .. autosummary::

       dispersion
       forward
       backward
       test
       predict_positions
    """

    def __init__(self, config, telescope=None):
        """
        Initialize spectrograph from optical configuration.

        :param OptConfig config: optical configuration
        :param Telescope telescope: input telescope if any
        """

        self.config = config                       #: Optical configuration

        self.telescope = telescope                 #: Telescope
        self.collimator = Collimator(self.config)  #: Collimator
        self.grism = Grism(self.config)            #: Grism
        self.camera = Camera(self.config)          #: Camera
        self.detector = Detector(self.config)      #: Detector

    @property
    def gamma(self):
        r"""
        Spectrograph magnification :math:`f_{\mathrm{cam}}/f_{\mathrm{coll}}`.
        """

        return self.camera.flength / self.collimator.flength

    def __str__(self):

        s = [" Spectrograph ".center(LINEWIDTH, '-')]
        if self.telescope:
            s.append(self.telescope.__str__())
        s.append(self.collimator.__str__())
        s.append(self.grism.__str__())
        s.append(self.camera.__str__())
        s.append(self.detector.__str__())
        s.append(f"Spectrograph magnification: {self.gamma:.3f}")
        wref = self.config.wref
        s.append("Central dispersion: "
                 f"{self.dispersion(wref)*1e10*self.detector.pxsize:.2f} AA/px "
                 f"at {wref*1e06:.2f} µm")

        return '\n'.join(s)

    def dispersion(self, wavelength, order=1, eps=1e-6):
        r"""
        Spectral dispersion (approximate) [m/m].

        This is given by :math:`D(\lambda) = (\mathrm{d}y /
        \mathrm{d}\lambda)^{-1}` with

        .. math:: y = f_{\mathrm{cam}}\tan\beta

        and

        .. math:: \sin\beta = m \rho \lambda - n(\lambda) \sin(A).

        :param float wavelength: wavelength [m]
        :param int order: dispersion order
        :return: spectral dispersion [m/m]
        """

        from scipy.misc import derivative

        def yoverf(l):

            sinbeta = (order * self.grism.grating.rho * 1e3 * l -
                       self.grism.prism.material.index(l) *
                       N.sin(self.grism.prism.angle))

            return N.tan(N.arcsin(sinbeta))

        dydl = self.camera.flength * derivative(yoverf, wavelength,
                                                dx=wavelength * eps)

        return 1 / dydl

    def forward(self, source, mode=1):
        """
        Forward light propagation from a focal-plane point source.

        :param source: input source
        :param mode: observing mode (dispersion order or photometric band)
        :return: (complex) 2D-positions in detector plane
        :rtype: :class:`numpy.ndarray`
        """

        wavelengths = source.coords['wavelength'].data

        # Telescope if any
        positions = (self.telescope.forward(source.zcoord)
                     if self.telescope else source.zcoord)
        # Collimator
        directions = self.collimator.forward(
            positions, wavelengths, self.gamma)
        if is_spectred(mode):   # Spectroscopic mode: grism
            directions = self.grism.forward(
                directions, wavelengths, order=mode)
        # Camera
        positions = self.camera.forward(directions, wavelengths)
        # Detector
        positions = self.detector.forward(positions)

        return positions

    def backward(self, position, wavelength, mode=1):
        """
        Backward light propagation from a detector-plane 2D-position
        and wavelength.

        :param complex position: 2D-position in the detector plane [m]
        :param float wavelength: wavelength [m]
        :param mode: observing mode (dispersion order or photometric band)
        :return: 2D-position in the focal plane [m]
        :rtype: complex
        """

        # Detector
        position = self.detector.backward(position)
        # Camera
        direction = self.camera.backward(position, wavelength)
        if is_spectred(mode):   # Spectroscopic mode
            # Grism
            direction = self.grism.backward(direction, wavelength, order=mode)
        # Collimator
        position = self.collimator.backward(direction, wavelength, self.gamma)
        # Telescope if any
        coords = (self.telescope.backward(position)
                  if self.telescope else position)

        return coords

    def test(self, waves=None, zcoord=(1e-3 + 2e-3j), mode=1, verbose=False):
        """
        Test forward and backward propagation in spectrograph.

        :param list waves: wavelength vector to be tested
        :param complex zcoord: input (complex) coordinates to be tested
        :param mode: observing mode (dispersion order or photometric band)
        :param bool verbose: verbose-mode
        :return: boolean result of the forward/backward test
        """

        # Test source
        if waves is None:       # Single test wavelength
            waves = N.array([self.config.wref])

        source = SourceXDA(waves, zcoord=zcoord)

        if verbose:
            print(f" SPECTROGRAPH TEST - MODE {mode} ".center(LINEWIDTH, '='))
            print("| Input source:", str_SourceXDA(source))
            print("| Wavelengths [µm]:", waves * 1e6)

        spectred = is_spectred(mode)  # Spectroscopic mode

        # Forward step-by-step ------------------------------

        # Telescope
        if self.telescope:
            fpositions = self.telescope.forward(source.zcoord)
            if verbose:
                print("> Positions (tel.forward) [×1e6]:", fpositions * 1e6)
        else:
            fpositions = source.zcoord

        # Collimator
        fdirections = self.collimator.forward(fpositions, waves, self.gamma)
        if verbose:
            print("> Directions (coll.forward) [×1e6]:", fdirections * 1e6)

        if spectred:            # Spectroscopic mode
            # Grism
            fdirections = self.grism.forward(fdirections, waves, order=mode)
            if verbose:
                print("> Directions (grism.forward) [×1e6]:", fdirections * 1e6)

        # Camera
        dpositions = self.camera.forward(fdirections, waves)
        if verbose:
            print("> Positions (camera.forward) [mm]:", dpositions * 1e3)

        # Detector
        dpositions = self.detector.forward(dpositions)
        if verbose:
            print("> Positions (detector) [mm]:", dpositions * 1e3)

        # Backward step-by-step ------------------------------

        output = []             # Round-trip positions, to be compared to input

        # Loop over positions in detector plane
        for lbda, dpos in zip(waves, N.atleast_1d(dpositions)):
            if verbose:
                print("| Test position (detector) [mm]:", dpos * 1e3,
                      "Wavelength [µm]:", lbda * 1e6)

            # Detector
            dpos = self.detector.backward(dpos)
            if verbose:
                print("< Direction (detector.backward) [mm]:", dpos * 1e3)

            # Camera
            bdirection = self.camera.backward(dpos, lbda)
            if verbose:
                print("< Direction (camera.backward) [×1e6]:", bdirection * 1e6)

            if spectred:        # Spectroscopic mode
                # Grism
                bdirection = self.grism.backward(bdirection, lbda, order=mode)
                if verbose:
                    print("< Direction (grism.backward) [×1e6]:",
                          bdirection * 1e6)

            # Collimator
            fposition = self.collimator.backward(bdirection,
                                                 lbda, self.gamma)

            # Telescope
            if self.telescope:
                tdirection = self.telescope.backward(fposition)
                if verbose:
                    print("< Position (coll.backward) [mm]:",
                          fposition * 1e3)
                    print("< Direction (tel.backward) [×1e6]:",
                          tdirection * 1e6,
                          "---OK---"
                          if N.isclose(tdirection, source.zcoord)
                          else '***ERROR***')

                output.append(tdirection)
            else:
                if verbose:
                    print("< Focal-plane position (backward) [mm]:",
                          fposition * 1e3,
                          "---OK---"
                          if N.isclose(fposition, source.zcoord)
                          else '***ERROR***')

                output.append(fposition)

        return N.allclose(source.zcoord, output)

    def predict_positions(self, simcfg, **kwargs):
        """
        Simulate detector spectra from optical model.

        :param SimConfig simcfg: input simulation configuration
        :param kwargs: configuration options (e.g. predicted `modes`)
        :return: predicted positions [m]
        :rtype: :class:`DetectorPositions`
        """

        # Update simulation configuration on-the-fly
        if kwargs:
            # Make a copy, to leave input simcfg intact
            simcfg = SimConfig(simcfg)
            simcfg.update(kwargs)

        # Input coordinates
        coords = simcfg.get_coordinates()      # 1D complex array

        # Input source (coordinates will be updated later on)
        waves = simcfg.get_wavelengths(self.config)

        # Initiate detector positions
        detector = DetectorPositions(waves, coords,
                                     spectrograph=self,
                                     name=self.config.name)

        # Simulated observing modes
        modes = simcfg.get('modes', [1])

        source = SourceXDA(waves)  # zcoord will be updated as needed
        def _update_zcoord(source, zcoord):
            source.attrs['zcoord'] = zcoord
            return source

        # Simulate forward propagation for all input coordinates
        for mode in modes:      # Loop over observing modes
            positions = N.transpose([
                self.forward(_update_zcoord(source, xy), mode=mode)
                for xy in coords ])  # ([nlbda,] npos)
            if positions.ndim == 1:  # No chromatic dimension
                # Expand along wavelength dimension
                positions = N.resize(positions, detector.xda.shape[1:])
            detector.add_mode(mode,
                              DetectorPositionsXDA(waves, coords, positions))

        return detector

    def update(self, **kwargs):
        """
        Update both optical configuration and structure parameters.
        """

        for name, value in kwargs.items():
            # Test parameter validity
            if name not in self.config:
                # raise KeyError(f"Unknown optical parameter {name!r}")
                warnings.warn(f"Unknown optical parameter {name!r}")
            # Update optical configuration
            self.config[name] = value
            # Update structure
            attrs = name.split('_')  # [ attributes ]
            obj = self
            for attr in attrs[:-1]:
                try:
                    obj = getattr(obj, attr)
                except AttributeError as error:
                    raise KeyError(
                        "Unknown attribute "
                        f"{attr!r} in {obj.__class__.__name__}") from error
            attr = attrs[-1]
            try:
                getattr(obj, attr)
            except AttributeError as error:
                raise KeyError(
                    "Unknown final attribute "
                    f"{attr!r} in {obj.__class__.__name__}") from error
            else:
                setattr(obj, attr, value)

    def optimize(self, positions, simcfg, modes=None,
                 optparams=['telescope_flength',
                            'collimator_flength',
                            'camera_flength'], tol=1e-6):
        """
        :mod:`scipy.optimize` optical parameters to match target detector
        positions, according to simulation configuration.

        :param DetectorPositions positions: target positions
        :param SimConfig simcfg: simulation configuration
        :param list modes: adjusted observing modes (default: simulated modes)
        :param list optparams: optical parameters to be adjusted
        :param float tol: optimization tolerance
        :return: result from the optimization
        :rtype: :class:`scipy.optimize.OptimizeResult`
        :raise KeyError: unknown optical parameter
        """

        import scipy.optimize as SO

        print(" SPECTROGRAPH ADJUSTMENT (scipy.optimize) ".center(LINEWIDTH, '='))

        # Simulation parameters
        if modes is None:
            modes = simcfg.get('modes', [1])
        print("Adjusted modes:", modes)

        try:
            guessparams = [ self.config[name] for name in optparams ]
        except KeyError as error:
            raise KeyError(f"Unknown optical parameter {name!r}") from error

        print(" Initial parameters ".center(LINEWIDTH, '-'))
        for name, value in zip(optparams, guessparams):
            print(f"  {name:20}: {value}")

        # Initial guess simulation
        mpositions = self.predict_positions(simcfg)
        # Test compatibility with objective detector positions only once
        mpositions.check_alignment(positions)

        rmss = []
        for mode in modes:
            rms = positions.compute_rms(mpositions, mode=mode)
            print(f"Mode {mode} RMS: {rms * 1e3} mm "
                  f"= {rms / self.detector.pxsize} px")
            rmss.append(rms)
        rms = (sum( rms ** 2 for rms in rmss ) / len(modes)) ** 0.5
        print(f"Total RMS: {rms * 1e3} mm = {rms / self.detector.pxsize} px")

        def objfun(params, positions):
            """Sum of squared mean offset position."""

            # Update optical configuration
            self.update(**dict(list(zip(optparams, params))))
            # Simulate
            mpositions = self.predict_positions(simcfg)
            dtot = sum( ((mpositions.xda.sel(mode=mode) -
                          positions.xda.sel(mode=mode)).abs()**2).values.mean()
                        for mode in modes )

            return dtot

        result = SO.minimize(objfun,
                             guessparams,
                             args=(positions,),
                             method='tnc',
                             options={'disp': True, 'xtol': tol})
        print(f"Minimization {'succeeded' if result.success else 'failed'}: "
              f"{result.message}")

        if result.success:
            print(" Adjusted parameters ".center(LINEWIDTH, '-'))
            for name, value in zip(optparams, N.atleast_1d(result.x)):
                print(f"  {name:20}: {value}")
            # Compute final RMS
            result.rms = (result.fun / len(modes)) ** 0.5  # [m]
            print(f"  RMS: {result.rms * 1e3} mm "
                  f"= {result.rms / self.detector.pxsize} px")

        return result

    def adjust(self, positions, simcfg, modes=None,
               optparams=['telescope_flength',
                          'collimator_flength',
                          'camera_flength'], **options):
        """
        :pypi:`iminuit` adjustement of optical parameters to match target
        detector positions, according to simulation configuration.

        :param DetectorPositions positions: target positions
        :param SimConfig simcfg: simulation configuration
        :param list modes: adjusted observing modes (default: simulated modes)
        :param list optparams: optical parameters to be adjusted
        :param dict options: iminuit additional options
        :return: result from the optimization
        :rtype: :class:`iminuit.Minuit`
        :raise KeyError: unknown optical parameter
        """

        from iminuit import Minuit

        print(" SPECTROGRAPH ADJUSTMENT (Minuit) ".center(LINEWIDTH, '='))

        # Simulation parameters
        if modes is None:
            modes = simcfg.get('modes', [1])
        print("Adjusted modes:", modes)

        try:
            guessparams = [self.config[name] for name in optparams]
        except KeyError as error:
            raise KeyError("Unknown optical parameter") from error

        kwargs = {'name': optparams}
        errors = {}

        print(" Initial parameters ".center(LINEWIDTH, '-'))
        for name, value in zip(optparams, guessparams):
            print(f"  {name:20}: {value}")
            kwargs[name] = value
            errors[name] = abs(value) if value else 1e-3

        kwargs.update(**options)

        # Initial guess simulation
        mpositions = self.predict_positions(simcfg)
        # Test compatibility with objective detector positions only once
        mpositions.check_alignment(positions)

        rmss = []
        for mode in modes:
            rms = positions.compute_rms(mpositions, mode=mode)
            print(f"Mode {mode} RMS: {rms * 1e3} mm "
                  f"= {rms / self.detector.pxsize} px")
            rmss.append(rms)
        rms = (sum(rms ** 2 for rms in rmss) / len(modes)) ** 0.5
        print(f"Total RMS:  {rms * 1e3} mm = {rms / self.detector.pxsize} px")

        def objfun(*args):
            """Sum of squared mean offset position."""

            # Update optical configuration
            self.update(**dict(list(zip(optparams, args))))
            # Simulate
            mpositions = self.predict_positions(simcfg)
            dtot = sum((N.abs(mpositions.xda.sel(mode=mode) -
                              positions.xda.sel(mode=mode))**2).values.mean()
                       for mode in modes)
            return dtot

        # Minuit optimizer
        objfun.errordef = Minuit.LEAST_SQUARES
        minuit = Minuit(objfun, **kwargs)
        for k, v in errors.items():
            minuit.errors[k] = v
        if minuit.print_level >= 1:
            print(minuit.params)

        # Optimization
        minuit.migrad()

        print("Minuit minimization: "
              f"{'succeeded' if minuit.valid else 'failed'}")

        if minuit.valid:
            print(" Adjusted parameters ".center(LINEWIDTH, '-'))
            for name in optparams:
                print(f"  {name:20}: "
                      f"{minuit.values[name]} ± {minuit.errors[name]}")
            # Update parameters to adjusted values
            objfun(*minuit.values)
            # Compute final RMS
            rms = (minuit.fval / len(modes)) ** 0.5  # [m]
            print(f"  RMS: {rms * 1e3} mm = {rms / self.detector.pxsize} px")

        return minuit

# Utility functions =======================================


def is_spectred(mode):
    """
    Is observational *mode* a spectroscopic (int-like) or photometric
    (str-like) mode?
    """

    # isinstance(mode, int) does not catch N.int64
    try:
        int(mode)
        return True
    except ValueError:
        return False


def str_mode(mode):
    """String 'Order #X' or 'Band Y' from *mode*."""

    return f"{'Order #' if is_spectred(mode) else 'Band '}{mode}"


def rect2pol(position):
    r"""
    Convert complex position(s) :math:`x + jy` into modulus :math:`r` and
    phase :math:`\phi`.

    :param complex position: 2D-position(s) :math:`x + jy`
    :return: (r, phi) [rad]
    """

    return N.absolute(position), N.angle(position)


def pol2rect(r, phi):
    r"""
    Convert modulus :math:`r` and phase :math:`\phi` into complex position(s)
    :math:`x + jy = r\exp(j\phi)`.

    :param float r: module(s)
    :param float phi: phase(s) [rad]
    :return: complex 2D-position(s)
    """

    return r * N.exp(1j * phi)


def dump_mpld3(fig, filename):
    """
    Dump figure to `mpld3 <http://mpld3.github.io/>`_ HTML-file.
    """

    import mpld3                # Raise ImportError if needed

    for ax in fig.axes:
        # Remove legend title if any
        # (https://github.com/jakevdp/mpld3/issues/275)
        leg = ax.get_legend()
        if leg is not None:
            txt = leg.get_title().get_text()
            if txt == 'None':
                leg.set_title("")

    # Add mouse position
    mpld3.plugins.connect(fig,
                          mpld3.plugins.MousePosition(fontsize='small'))
    # Save figure to HTML
    mpld3.save_html(fig, filename,
                    no_extras=False, template_type='simple')
    print("MPLD3 figure saved in", filename)



# -*- coding: utf-8 -*-
# Time-stamp: <2021-04-19 00:17 ycopin@lyonovae03>

import astropy.modeling as AM

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>"


def fit_legendre2D(x, y, z, deg=(2, 2)):
    """
    Fit a 2D-polynomial to *z(x, y)*.
    """

    # Model: 2D Legendre polynomials
    p_init = AM.models.Legendre2D(*deg)

    # Least-square fit
    fitter = AM.fitting.LinearLSQFitter()
    p_fit = fitter(p_init, x, y, z)

    # Compute and store RMS
    p_fit.rms = (fitter.fit_info['residuals'] / z.size) ** 0.5

    return p_fit                # 2D-polynomial model



# -*- coding: utf-8 -*-
# Time-stamp: <2021-04-18 23:27 ycopin@lyonovae03>

"""
mcmc
----

Monte-Carlo Markov Chain methods.

.. autosummary::

   run_mcmc
   plot_mcmc_chains
   plot_mcmc_corner
   mcmc_best_params
"""

import numpy as N
import matplotlib.pyplot as P

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>"


def run_mcmc(spectro, positions, simcfg, optparams, lnprior, modes=None,
             nwalkers=10, nsteps=500, outfile='chains.dat'):
    """
    Monte-Carlo Markov Chain exploration of parameter space, using
    :pypi:`emcee`.

    .. Warning:: very preliminary implementation

    :param Spectrograph spectro: spectrograph
    :param DetectorPositions positions: target positions
    :param SimConfig simcfg: simulation configuration
    :param list optparams: optical parameters to be probed
    :param function lnprior: log-likelihood prior function
    :param list modes: adjusted observing modes (default: simulated modes)
    :param int nwalkers: the number of walkers will be
                         `2*len(optparams)*nwalkers`
    :param int nsteps: number of MCMC-steps
    :param str outfile: incremental file output
    :return: Monte-Carlo Markov Chains array (nwalkers, nsteps, ndim)
    :rtype: :attr:`emcee.EnsembleSampler.chain`
    :raise KeyError: unknown optical parameter

    **Reference:** `Foreman-Mackey et al. 2012
    <http://adsabs.harvard.edu/abs/2013PASP..125..306F>`_
    """

    import emcee

    # Simulation parameters
    if modes is None:
        modes = simcfg.get('modes', [1])
    print("Adjusted modes:", modes)

    try:
        guessparams = N.array([ spectro.config[name] for name in optparams ])
    except KeyError as error:
        raise KeyError(f"Unknown optical parameter {name!r}") from error

    print("Initial parameters:")
    for name, value in zip(optparams, guessparams):
        print(f"  {name:20}: {value}")

    # Initial guess simulation
    mpositions = spectro.predict_positions(simcfg)
    # Test compatibility with objective detector positions only once
    mpositions.check_alignment(positions)

    rmss = []
    for mode in modes:
        rms = positions.compute_rms(mpositions, mode=mode)
        print(f"Mode {mode} RMS: {rms / 0.001} mm = "
              f"{rms / spectro.detector.pxsize} px")
        rmss.append(rms)
    rms = (sum( rms ** 2 for rms in rmss ) / len(modes)) ** 0.5
    print(f"Total RMS:  {rms / 0.001} mm = {rms / spectro.detector.pxsize} px")

    def lnlike(params):
        """Log-likelihood function (without prior)."""

        # Update optical configuration
        spectro.update(**dict(list(zip(optparams, params))))
        # Simulate
        mpositions = spectro.predict_positions(simcfg)
        dtot = sum(
            ((mpositions[mode] - positions[mode]).abs()**2).values.mean()
            for mode in modes )

        return -0.5 * (dtot / len(modes))**0.5

    def lnprob(params):
        """Full log-likelihood function, including prior."""

        prior = lnprior(params)
        if N.isfinite(prior):
            prior += lnlike(params)

        return prior

    ndim = len(optparams)
    nwalkers *= 2 * ndim

    print(f"MCMC sampler: ndim={ndim}, nwalkers={nwalkers}, nsteps={nsteps}")

    # Emcee sampler
    sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob)

    # Sampling
    initparams = N.array([ guessparams + 1e-4 * N.random.randn(ndim)
                           for i in range(nwalkers)])

    if outfile:
        f = open(outfile, "w")
        f.close()

    for i, (pos, lnp, state) in enumerate(sampler.sample(initparams,
                                                         iterations=nsteps)):
        if (i+1) % 10 == 0:
            print(f"{100.0 * i / nsteps:.1f}%")

        if outfile:
            f = open(outfile, "a")
            for k in range(pos.shape[0]):
                f.write(f"{k:5d}  {' '.join(map(str, pos[k])):s}\n")
            f.close()

    return sampler.chain    # MCMC chains (nwalkers, nsteps, ndim)


def plot_mcmc_chains(chains, parameters):
    """Plot MCMC-chains (nwalkers, nsteps, ndim)."""

    assert chains.ndim == 3 and chains.shape[2] == len(parameters)

    fig, axs = P.subplots(len(parameters), 1, sharex=True)
    for i, (ax, name) in enumerate(zip(axs, parameters)):
        ax.plot(chains[:, :, i].T, color='k', alpha=0.4)
        ax.set_title(name)
    fig.tight_layout()

    return fig


def plot_mcmc_corner(chains, parameters, burnin=0):
    """
    :pypi:`corner` plot of MCMC parameters.

    **Reference:** `Foreman-Mackey 2016
    <http://joss.theoj.org/papers/10.21105/joss.00024>`_
    """

    import corner

    assert chains.ndim == 3 and chains.shape[2] == len(parameters)
    assert burnin < chains.shape[1]

    # Remove 1st burnin steps, and reshape (nwalkers*nsteps, ndim)
    samples = chains[:, burnin:, :].reshape((-1, len(parameters)))

    fig = corner.corner(samples, labels=parameters)

    return fig


def mcmc_best_params(chains, parameters, burnin=0):
    """
    Compute best parameter estimates and assymetric 1-sigma errors from
    marginalized distributions.  Return `{name: (median, low_err, high_err)}`.
    """

    assert chains.ndim == 3 and chains.shape[2] == len(parameters)
    assert burnin < chains.shape[1]

    # Remove 1st burnin steps, and reshape (nwalkers*nsteps, ndim)
    samples = chains[:, burnin:, :].reshape((-1, len(parameters)))

    # Compute percentiles (nperc, ndim)
    percentiles = N.percentile(samples, [15.9, 50, 84.1], axis=0)  # (3, ndim)

    bests = { name: (median, low, high)
              for name, median, low, high in zip(
                  parameters,
                  percentiles[1],  # Median
                  percentiles[1] - percentiles[0],   # Median - low
                  percentiles[2] - percentiles[1])}  # High - median

    print("MCMC best estimates:")
    for name in parameters:
        median, low, high = bests[name]
        print(f"  {name:20s} = {median:+g} + {low:g} - {high:g}")

    return bests



# -*- coding: utf-8 -*-
# Time-stamp: <2021-05-11 01:26 ycopin@lyonovae03>

"""
nisp
----

NISP-specific tools, including a :class:`ZemaxPositions` handler.

.. autosummary::

   ZemaxPositions
"""

import os
import sys
import warnings

import numpy as N
import matplotlib.pyplot as P
import pandas as PD
import xarray as X

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>"

import spectrogrism.spectrogrism as S

#: NISP effective optical configuration, R-grism
#:
#: .. Note:: The detector plane is tiled with 4×4 detectors of 2k×2k pixels of
#:    18 µm; the spectrograph has a mean magnification (`NISPPlateScale`) of
#:    0.5 approximately.  Hence a focal plane of approximately 29×29 cm².
NISP_R = S.OptConfig([
    ('name', "NISP-R"),                   # Configuration name
    ('wave_ref', 1.5e-6),                 # Reference wavelength [m]
    ('wave_range', [1.25e-6, 1.85e-6]),   # Wavelength range [m]
    # Telescope
    ('telescope_flength', 24.5),          # Telescope focal length [m]
    # Grism
    ('grism_dispersion', 9.8),            # Rough spectral dispersion [AA/px]
    ('grism_prism_material', 'FS'),       # Prism glass
    ('grism_grating_material', 'FS'),     # Grating resine
    ('grism_prism_angle', 2.08 / S.RAD2DEG),   # Prism angle [rad]
    # ('grism_grating_rho', 19.29),       # Grating groove density [lines/mm]
    ('grism_grating_rho', 13.72),         # Grating groove density [lines/mm]
    ('grism_grating_blaze', 2.6 / S.RAD2DEG),  # Blaze angle [rad]
    # Detector
    ('detector_pxsize', 18e-6),           # Detector pixel size [m]
])

# Guessed values (not from official documents)
NISP_R.update([
    # Telescope ------------------------------
    # Collimator ------------------------------
    ('collimator_flength', 1924e-3),  # Focal length [m]
    ('collimator_gdist_x0', -0.31e-3),    # Distortion offset
    ('collimator_gdist_y0', 54.6e-3),  # Distortion offset
    ('collimator_gdist_K1', 447e-3),  # Radial distortion
    # ('collimator_gdist_P1', +5.4e-5), ('collimator_gdist_P2', -7.2e-5),
    # Grism ------------------------------
    ('grism_prism_angle', 2.078 / S.RAD2DEG),   # Prism angle [rad]
    ('grism_grating_rho', 13.09),  # Grating groove density [lines/mm]
    ('grism_prism_tiltx', -362. / S.RAD2MIN),  # Prism x-tilt (around apex/groove axis) [rad]
    ('grism_prism_tilty', +22.6 / S.RAD2MIN),  # Prism y-tilt [rad]
    ('grism_prism_tiltz', +7.59 / S.RAD2MIN),  # Prism z-tilt (around optical axis) [rad]
    # Camera ------------------------------
    ('camera_flength', 975.5e-3),               # Focal length [m]
    ('camera_gdist_x0', 1.8e-3),
    ('camera_gdist_y0', 218.5e-3),
    ('camera_gdist_K1', -62.5e-3),
    # ('camera_gdist_P1', 0e-5), ('camera_gdist_P2', -1.0e-5),
    # Detector ------------------------------
    # Without input recentering
    # ('detector_dx', +0.70e-3),                 # Detector x-offset [m]
    # ('detector_dy', +179.7e-3),                # Detector y-offset [m]
    # With input offset of -0.85 deg
    ('detector_dx', +656e-6),           # Detector x-offset [m]
    ('detector_dy', +880e-6),           # Detector y-offset [m]
])

NISP_Y = S.OptConfig([
    ('name', "NISP-Y"),                   # Configuration name
    ('wave_range', [952e-9, 1214e-9]),    # Wavelength range [m]
])


class ZemaxPositions(S.DetectorPositions):
    """
    Zemax simulated positions, in spectroscopic or photometric mode.

    Zemax configuration modes ('confNB'):

    * 1: B-grism (NISP-S)
    * 2, 3, 4: R-grisms (NISP-S)
    * 5, 6, 7: Y, J, H (NISP-P)
    """

    colnames = """
confNb wave xindeg yindeg ximgmm yimgmm pxsize nxpup nypup nximg nyimg
ximgcpx yimgcpx ximgcmm yimgcmm xpsfcmm ypsfcmm
ee50mm ee80mm ee90mm ellpsf papsfdeg""".split()  #: Input column names

    def __init__(self, simulations, colname='psfcmm'):
        """
        Initialize from `simulations` = {mode: filename}.
        """

        initialized = False

        # Extract simulation name from input dictionary
        name = simulations.get("name", "Anonymous")

        self.filenames = simulations  # {mode: filename}

        for mode, filename in self.filenames.items():
            if mode == 'name':
                continue
            # Read dataframe (wavelengths as index, input coords in columns)
            df = self.read_simulation(filename=filename, colname=colname)
            if not initialized:
                # Initialize DetectorPositions from wavelengths and coordinates
                super().__init__(
                    wavelengths=df.index.values,
                    coordinates=df.columns.values.astype(complex),
                    name=name)
                initialized = True
            # Add dataframe
            self.add_mode(mode, X.DataArray(df))

    @property
    def orders(self):
        """Dispersion orders (int modes)."""

        return [ mode for mode in self.modes if S.is_spectred(mode) ]

    @property
    def bands(self):
        """Undispersed photometric bands (string modes)."""

        return [ mode for mode in self.modes if not S.is_spectred(mode) ]

    def __str__(self):

        s = f"Simulations {self.name!r}: {len(self.modes)} mode(s)"
        for order in self.orders:
            s += f"\n  Order #{order}: {self.filenames[order]}"
        for band in self.bands:
            s += f"\n  Band   {band}: {self.filenames[band]}"

        waves = self.wavelengths * 1e6  # [µm]
        coords = self.coordinates
        s += (f"\n  Wavelengths: {len(waves)} steps "
              f"from {min(waves):.2f} to {max(waves):.2f} µm")
        s += f"\n  Coords: {len(coords)} sources"

        return s

    @classmethod
    def read_simulation(cls, filename, colname='psfcmm'):
        """
        Read single-mode simulation.

        :param str filename: input simulation filename
        :param str colname: column name suffix to be used as output positions
                            (in **mm**)
        :return: (wavelength, input coordinates) dataframe
        :rtype: :class:`pandas.DataFrame`

        .. Warning:: this function relies on :meth:`pandas.DataFrame.pivot`.
           Because of Pandas `issue #12666
           <https://github.com/pydata/pandas/issues/12666>`_ and related Numpy
           `issue 7535 <https://github.com/numpy/numpy/issues/7435>`_, this
           requires Numpy >= v1.11.
        """

        if 'mm' not in colname:  # Input coords are supposed to be in mm
            raise NotImplementedError(
                "Only output columns in mm are supported.")

        data = N.genfromtxt(filename, dtype=None, names=cls.colnames)

        # Cleanup: some Xin are arbitrarily close to zero
        warnings.warn("Setting approximately null xindeg to 0")
        data['xindeg'][N.abs(data['xindeg']) < 1e-12] = 0

        # Cleanup: offset yindeg by -0.85 deg
        warnings.warn("Offsetting Yin by -0.85 deg")
        data['yindeg'] += -0.85

        # Cleanup: upper-right position has no 1.85 µm wavelength
        warnings.warn("Discarding wavelengths > 1.81 µm")
        data = data[data['wave'] < 1.81]

        # Convert to DataFrame and modify/add columns
        df = PD.DataFrame(data)
        # Wavelengths [m]
        df['wave'] /= 1e6
        # Output complex coordinates [m]
        df['xyoutm'] = (df['x' + colname] + 1j * df['y' + colname]) / 1e3

        # Convert input coordinates [deg] into 'xyinrad' columns [rad]
        pivot = df.pivot(index="wave",
                         columns=["xindeg", "yindeg"], values="xyoutm")
        pivot.columns = PD.Index([ complex(x, y) / S.RAD2DEG
                                   for x, y in pivot.columns ],
                                 name='xyinrad')

        # Match DetectorPositions names
        pivot.rename_axis(index='wavelength',
                          columns='coordinate', inplace=True)

        # Make sure wavelengths and input complex coordinates are sorted
        pivot.sort_index(inplace=True)
        pivot = pivot.reindex(             # Complex coordinates
            N.sort(pivot.columns.values.astype(complex)), axis=1)

        return pivot

    def get_simcfg(self):
        """
        Generate a :class:`spectrogrism.spectrogrism.SimConfig`
        corresponding to current simulation.
        """

        # Wavelengths [m]
        waves = self.wavelengths
        # Input (complex) coordinates [rad]
        coords = self.coordinates
        # Convert back to [[x, y]]
        coords = N.vstack((coords.real, coords.imag)).T

        simcfg = S.Configuration([('name', self.name),
                                  ('wave_npx', len(waves)),
                                  ('wave_range', [min(waves), max(waves)]),
                                  ('modes', self.modes),
                                  ('input_coords', coords)
                                  ])

        return S.SimConfig(simcfg)

    def plot_input(self, ax=None, **kwargs):
        """Plot input coordinates [deg]."""

        if ax is None:
            fig = P.figure()
            ax = fig.add_subplot(
                1, 1, 1,
                xlabel="x [deg]", ylabel="y [deg]",
                title=f"Simulation '{self.name}'\nInput positions")

        coords = self.coordinates * S.RAD2DEG  # [deg]

        ax.scatter(coords.real, coords.imag, **kwargs)

        ax.set_aspect('equal', adjustable='datalim')

        return ax

    def plot_output(self, ax=None, modes=None, subsampling=0, **kwargs):
        """Plot output (detector-level) positions [mm]."""

        if modes is None:
            modes = self.modes

        ax = self.plot(ax=ax, modes=modes,
                       subsampling=subsampling, **kwargs)

        title = f"Simulation '{self.name}'"
        if subsampling:
            title += f" (subsampled ×{subsampling})"
        ax.set_title(title)

        return ax

    def plot_offsets(self, other, ax=None, mode=1, nwaves=3):
        """Plot output (detector-level) coordinate offsets [px]."""

        if ax is None:
            fig = P.figure()
            ax = fig.add_subplot(1, 1, 1,
                                 xlabel="x [mm]", ylabel="y [mm]")
            title = (f"Simulation {self.name!r}, {S.str_mode(mode)}\n"
                     "position offsets [px]")
            ax.set_title(title)
            ax.set_aspect('equal', adjustable='datalim')
        else:
            fig = None          # Will serve as a flag

        # Position offset
        zpos = self.xda.sel(mode=mode) * 1e3    # [mm]
        opos = other.xda.sel(mode=mode) * 1e3   # [mm]
        dpos = opos - zpos                      # [mm]
        dpos /= other.spectrograph.detector.pxsize * 1e3  # [px]

        # Create subsampled wavelength vector of length nwaves
        waves = self.wavelengths
        iwaves = N.linspace(0, len(waves) - 1, nwaves).round().astype(int)
        waves = waves[iwaves]
        # Generate colormap
        colorMap = P.matplotlib.cm.ScalarMappable(
            norm=P.matplotlib.colors.Normalize(vmin=waves[0], vmax=waves[-1]),
            cmap=P.get_cmap('Spectral_r'))   # (blue to red)

        qkey = None                       # Quiver legend (only once)

        for wave in waves:
            # Input (Zemax) position
            x, y = zpos.loc[wave].real, zpos.loc[wave].imag    # [mm]
            # Offset between modeled and input positions
            dx, dy = dpos.loc[wave].real, dpos.loc[wave].imag  # [px]

            col = colorMap.to_rgba(wave)                       # Get color
            q = ax.quiver(x, y, dx, dy, color=col, units='width')
            if qkey is None:              # Add quiver label
                qkey = ax.quiverkey(q, 0.9, 0.95, 10, "10 px",
                                    labelpos='E', coordinates='figure')

        ax.axis([-100, +100, -100, +100])               # [mm]

        return ax


# Main ====================================================

if __name__ == '__main__':

    try:
        import seaborn
        seaborn.set_style("darkgrid")
    except ImportError:
        pass

    subsampling = 3          # Subsample output plot
    adjust = [               # Adjusted optical parameters (all modes)
        # # 'telescope_gdist_K1', 'telescope_gdist_y0',
        # 'collimator_flength',
        # 'collimator_gdist_x0', 'collimator_gdist_y0', 'collimator_gdist_K1',
        # # 'collimator_gdist_P1', 'collimator_gdist_P2',
        # 'camera_flength',
        # 'camera_gdist_x0', 'camera_gdist_y0', 'camera_gdist_K1',
        # # 'camera_gdist_P1', 'camera_gdist_P2',
        # 'detector_dx', 'detector_dy',
    ]
    adjust_phot = [          # Adjusted photometric optical parameters
        # 'telescope_gdist_K1', 'telescope_gdist_y0',
        # 'collimator_gdist_x0', 'collimator_gdist_y0', 'collimator_gdist_K1',
        # 'camera_gdist_x0', 'camera_gdist_y0', 'camera_gdist_K1',
        # 'detector_dx', 'detector_dy',
    ]
    adjust_spec = [          # Adjust spectroscopic parameters
        # 'telescope_gdist_K1', 'telescope_gdist_y0',
        # 'collimator_gdist_x0', 'collimator_gdist_y0', 'collimator_gdist_K1',
        # 'collimator_cdist_C1',
        # 'grism_prism_angle', 'grism_grating_rho',
        # 'grism_prism_tiltx', 'grism_prism_tilty', 'grism_prism_tiltz',
        # 'camera_gdist_x0', 'camera_gdist_y0',
        # 'camera_gdist_K1',
        # 'camera_cdist_C1',
        # 'detector_dx', 'detector_dy',
    ]
    mcmc_params = [                    # MCMC exploration
        # 'collimator_gdist_x0', 'collimator_gdist_y0',
        # 'collimator_gdist_K1',
        # 'camera_gdist_x0', 'camera_gdist_y0',
        # 'camera_gdist_K1',
        # 'detector_dx', 'detector_dy',
    ]

    datadir = os.path.join(os.path.dirname(__file__), 'data')
    embed_html = False          # Generate MPLD3 figure
    plot_offset = True          # Offset plots
    load_config = f'{datadir}/nisp_adjusted2.yml' and False
    save_config = f'{datadir}/nisp_adjusted2.yml' and False

    # Zemax simulations
    simulations = S.Configuration([
        ("name", "Zemax"),
        (1, f"{datadir}/run_190315.dat"),            # 1st-order dispersed simulation
        # (0, f"{datadir}/run_011115_conf2_o0.dat"),   # 0th-order dispersed simulation
        # (2, f"{datadir}/run_161115_conf2_o2.dat"),   # 2nd-order dispersed simulation
        ('J', f"{datadir}/run_071215_conf6_J.dat"),  # J-band undispersed simulation
    ])
    print(simulations)

    zmx_pos = ZemaxPositions(simulations)
    print(zmx_pos)

    if '--plot' in sys.argv:        # Plot input positions
        ax = zmx_pos.plot_input()

    # Optical model
    if load_config:
        optcfg = S.OptConfig.load(load_config)
    else:
        optcfg = NISP_R             # Optical configuration (default NISP)

    simcfg = zmx_pos.get_simcfg()   # Simulation configuration

    spectro = S.Spectrograph(optcfg,
                             telescope=S.Telescope(optcfg))
    print(spectro)

    # Round-trip coherence test =============================

    print(" Spectrograph round-trip test ".center(S.LINEWIDTH, '-'))
    for mode in zmx_pos.modes:                  # Loop over all observing modes
        if not spectro.test(
                waves=zmx_pos.wavelengths,      # Test on all input wavelengths
                zcoord=zmx_pos.coordinates[0],  # Test on 1st input coordinates
                mode=mode, verbose=False):
            warnings.warn(f"{S.str_mode(mode)}: "
                          "forward and backward models do not match.")
        else:
            print(f"{S.str_mode(mode)}: OK")

    # Predictions and adjustments ==============================

    # Input Zemax positions
    kwargs = dict(s=20, edgecolor='k', linewidths=1)  # Outlined symbols
    ax = zmx_pos.plot_output(modes=zmx_pos.modes, **kwargs)

    if adjust:
        print('Optical adjustment on *all* modes')
        minuit = spectro.adjust(zmx_pos, simcfg,
                                modes=zmx_pos.modes, optparams=adjust)

    drms = {}

    # Imagery modes ------------------------------

    if adjust_phot and zmx_pos.bands:
        print('Optical adjustment on photometric modes')
        minuit = spectro.adjust(zmx_pos, simcfg,
                                modes=zmx_pos.bands, optparams=adjust_phot)

    # Predict positions on photometric modes
    phot_pos = spectro.predict_positions(simcfg, modes=zmx_pos.bands)
    phot_pos.check_alignment(zmx_pos)

    kwargs = {}                 # Default
    for band in zmx_pos.bands:  # Loop over photometric bands
        # Compute RMS on positions
        drms[band] = zmx_pos.compute_rms(phot_pos, mode=band)
        print(f"Band   {band}: RMS = {drms[band]*1e3:.4f} mm = "
              f"{drms[band] / spectro.detector.pxsize:.2f} px")
        phot_pos.plot(ax=ax, zorder=0,  # Draw below Zemax
                      modes=(band,),
                      label="{} {} (RMS={:.1f} px)".format(phot_pos.name, band,
                          drms[band] / spectro.detector.pxsize),
                      **kwargs)

    ax.axis([-100, +100, -100, +100])               # [mm]
    ax.set_aspect('equal', adjustable='datalim')
    ax.legend(loc='upper left', fontsize='small', frameon=True, framealpha=0.5)

    if plot_offset and zmx_pos.bands:             # Residual quiver plots
        fig, axs = P.subplots(1, len(zmx_pos.bands), squeeze=False)
        for ax, band in zip(axs.ravel(), zmx_pos.bands):
            zmx_pos.plot_offsets(phot_pos, ax=ax, mode=band)
            ax.set_aspect('equal', adjustable='box')
            ax.set_title(f"Band {band} "
                         f"(RMS={drms[band] / spectro.detector.pxsize:.1f} px)")

    # Spectroscopic modes ------------------------------

    # Input Zemax positions
    kwargs = dict(s=20, edgecolor='k', linewidths=1)  # Outlined symbols
    ax = zmx_pos.plot_output(modes=zmx_pos.orders,
                             subsampling=subsampling, **kwargs)

    if adjust_spec and zmx_pos.orders:
        print('Optical adjustment on spectroscopic modes')
        minuit = spectro.adjust(zmx_pos, simcfg,
                                modes=zmx_pos.orders, optparams=adjust_spec)

    # Predict positions on spectroscopic modes
    spec_pos = spectro.predict_positions(simcfg, modes=zmx_pos.orders)
    spec_pos.check_alignment(zmx_pos)

    # kwargs = dict(edgecolor=None, facecolor='none', linewidths=1)  # Open symbols
    kwargs = {}                   # Default
    for order in zmx_pos.orders:  # Loop over spectroscopic dispersion orders
        # Compute RMS on current order positions
        drms[order] = zmx_pos.compute_rms(spec_pos, mode=order)
        print(f"Order #{order}: RMS = {drms[order]*1e3:.4f} mm = "
              f"{drms[order] / spectro.detector.pxsize:.2f} px")
        spec_pos.plot(ax=ax, zorder=0,  # Draw below Zemax
                      modes=(order,), blaze=(order != 1),
                      subsampling=subsampling,
                      label=(f"{spec_pos.name} #{order} (RMS="
                             f"{drms[order]/spectro.detector.pxsize:.1f} px)"),
                      **kwargs)

    ax.axis([-100, +100, -100, +100])               # [mm]
    ax.set_aspect('equal', adjustable='datalim')
    ax.legend(loc='upper left', fontsize='small', frameon=True, framealpha=0.5)

    if embed_html:
        figname = zmx_pos.name + '_mpld3.html'
        try:
            S.dump_mpld3(ax.figure, figname)
        except ImportError:
            warnings.warn("MPLD3 is not available, cannot export to HTML.")

    if plot_offset:             # Residual quiver plots
        fig, axs = P.subplots(1, len(zmx_pos.orders), squeeze=False)
        for ax, order in zip(axs.ravel(), zmx_pos.orders):
            zmx_pos.plot_offsets(spec_pos, ax=ax, mode=order)
            ax.set_aspect('equal', adjustable='box')
            ax.set_title(f"Order #{order} "
                         f"(RMS={drms[order]/spectro.detector.pxsize:.1f} px)")

    # Save final configuration
    if save_config:
        optcfg.save(save_config)

    # MCMC sampling
    if mcmc_params:
        import pickle as pickle
        from spectrogrism import mcmc

        chain_name = "data/chains3"
        pklname = chain_name + '.pkl'
        burnin = 200

        if False:
            print(" SPECTROGRAPH EXPLORATION (MCMC) ".center(S.LINEWIDTH, '='))

            assert mcmc_params == [
                'collimator_gdist_x0', 'collimator_gdist_y0',
                'collimator_gdist_K1',
                'camera_gdist_x0', 'camera_gdist_y0',
                'camera_gdist_K1',
                'detector_dx', 'detector_dy',
            ]

            def lnprior_flat(params):
                """Hard-coded flat prior."""

                limits = [
                    (-1e-2, +1e-2),  # collimator_gdist_x0
                    (-0.1, 0.2),     # collimator_gdist_y0
                    (0.3, 0.6),      # collimator_gdist_K1
                    (-1e-2, +1e-2),  # camera_gdist_x0
                    (0.1, 0.35),     # camera_gdist_y0
                    (-0.2, 0.1),     # camera_gdist_K1
                    (-2e-3, +4e-3),  # detector_dx
                    (-2e-3, +4e-3),  # detector_dy
                ]

                for value, (vmin, vmax) in zip(params, limits):
                    if not vmin < value < vmax:
                        return -N.inf

                return 0

            chains = mcmc.run_mcmc(spectro, zmx_pos, simcfg,
                                   mcmc_params, lnprior_flat,
                                   modes=zmx_pos.modes)
            pickle.dump(chains, open(pklname, 'w'), protocol=-1)
            print("MCMC chains saved in", pklname)
        else:
            chains = pickle.load(open(pklname, 'r'))
            print("MCMC chains loaded from", pklname)

        fig = mcmc.plot_mcmc_chains(chains, mcmc_params)
        fig = mcmc.plot_mcmc_corner(chains, mcmc_params, burnin=burnin)
        mcmc.mcmc_best_params(chains, mcmc_params, burnin=burnin)

    P.show()



# -*- coding: utf-8 -*-
# Time-stamp: <2021-05-04 18:38 ycopin@lyonovae03>

"""
snifs
-----

SNIFS optical configuration and utilities.
"""

import warnings

import numpy as N

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>"

import spectrogrism.spectrogrism as S

#: SNIFS optical configuration, R-channel
SNIFS_R = S.OptConfig([
    ('name', "SNIFS-R"),                # Configuration name
    ('wave_ref', 0.76e-6),              # Reference wavelength [m]
    ('wave_range', [0.5e-6, 1.0e-6]),   # Standard wavelength range [m]
    # Telescope
    ('telescope_flength', 22.5),        # Focal length [m]
    # Collimator
    ('collimator_flength', 169.549e-3),    # Focal length [m]
    ('collimator_distortion', +2.141),     # r² distortion coefficient
    ('collimator_lcolor_coeffs', [-4.39879e-6, 8.91241e-10, -1.82941e-13]),
    # Grism
    ('grism_prism_material', 'BK7'),           # Prism glass
    ('grism_prism_angle', 17.28 / S.RAD2DEG),  # Prism angle [rad]
    ('grism_grating_rho', 200.),   # Grating groove density [lines/mm]
    ('grism_dispersion', 2.86),    # Informative spectral dispersion [AA/px]
    ('grism_grating_material', 'EPR'),         # Grating resine
    ('grism_grating_blaze', 15. / S.RAD2DEG),  # Blaze angle [rad]
    # Camera
    ('camera_flength', 228.014e-3),            # Focal length [m]
    ('camera_distortion', -0.276),             # r² distortion coefficient
    ('camera_lcolor_coeffs', [+2.66486e-6, -5.52303e-10, 1.1365e-13]),
    # Detector
    ('detector_pxsize', 15e-6),          # Detector pixel size [m]
    ('detector_angle', 0. / S.RAD2DEG),  # Rotation of the detector (0=blue is up)
])

#: SNIFS simulation configuration
SNIFS_SIMU = S.SimConfig([
    ('name', "standard"),                # Configuration name
    ('wave_npx', 6),                     # Nb of pixels per spectrum
    ('modes', (1, 0, 2, -1)),            # Dispersion orders
    # Focal plane sampling
    ('input_coords', N.linspace(-1e-2, 1e-2, 5)),  # Focal plane grid [m]
    ('input_angle', -10. / S.RAD2DEG),             # Focal plane angle [rad]
])


# Simulations ==============================


def plot_SNIFS(optcfg=SNIFS_R, simcfg=SNIFS_SIMU,
               test=True, verbose=False):
    """
    Test-case w/ SNIFS-like configuration.
    """

    # Optical configuration
    print(optcfg)

    # Spectrograph
    spectro = S.Spectrograph(optcfg)
    print(spectro)

    # Simulation configuration
    print(simcfg)

    if test:
        print(" Spectrograph round-trip test ".center(S.LINEWIDTH, '-'))
        for mode in simcfg['modes']:
            if not spectro.test(simcfg.get_wavelengths(optcfg),
                                mode=mode, verbose=verbose):
                warnings.warn(
                    f"Order #{mode}: backward modeling does not match.")
            else:
                print(f"Order #{mode:+d}: OK")

    positions = spectro.predict_positions(simcfg)
    ax = positions.plot(modes=(-1, 0, 1, 2), blaze=True)
    ax.legend(loc='upper left', fontsize='small', frameon=True, framealpha=0.5)
    ax.set_aspect('auto')
    ax.axis(N.array([-2000, 2000, -4000, 4000]) *
            spectro.detector.pxsize * 1e3)  # [mm]

    return ax


# Main ====================================================

if __name__ == '__main__':

    import matplotlib.pyplot as P
    try:
        import seaborn
        seaborn.set_style("whitegrid")
    except ImportError:
        pass

    ax = plot_SNIFS(test=True, verbose=True)

    embed_html = False
    if embed_html == 'mpld3':
        try:
            S.dump_mpld3(ax.figure, 'SNIFS-R_mpld3.html')
        except ImportError:
            warnings.warn("MPLD3 is not available, cannot export to HTML.")
    elif embed_html == 'bokeh':
        try:
            S.dump_bokeh(ax.figure, 'SNIFS-R_bokeh.html')
        except ImportError:
            warnings.warn("Bokeh is not available, cannot export to HTML.")
    elif embed_html:
        warnings.warn(f"Unknown HTML method '{embed_html}'")

    P.show()



# -*- coding: utf-8 -*-
# Time-stamp: <2021-05-04 18:21 ycopin@lyonovae03>

"""
distortion
-----------

Distortion utilities.

.. autosummary::

   StructuredGrid
   GeometricDistortion
   ChromaticDistortion
"""

import warnings

import numpy as N

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>"


class StructuredGrid:
    """
    A structured 2D-grid, stored as a complex 2D-array.

    .. autosummary::

       create
       reorder
       plot
       estimate_parameters
       rms
       adjust_distortion
       plot_offsets
    """

    def __init__(self, xy):

        self.xy = N.atleast_2d(xy).astype(complex)
        try:
            ny, nx = self.xy.shape
        except ValueError as error:
            raise ValueError(f"Invalid array shape {self.xy.shape}") from error

    def __str__(self):

        try:
            sign = self.signature
        except ValueError:
            sign = "unsigned"

        s = (f"Structured grid: {self.nx} × {self.ny} = "
             f"{self.nx * self.ny} positions, {sign}")

        return s

    @classmethod
    def create(cls, nx, ny, step='auto', offset=0, rotation=0.):
        """
        Create a regular rectangular grid.

        :param int nx: number of steps in *x*
        :param int ny: number of steps in *y*
        :param step: fixed (float) or automatic ('auto') stepsize
        :param complex offset: grid offset
        :param float rotation: grid angle [rad]
        """

        assert (nx, ny) >= (2, 2), "Invalid grid size."
        if step == 'auto':     # Grid extent will be ~ ± 1
            step = 2 / (nx * ny)**0.5

        x = N.arange(nx) - (nx - 1)/2.
        y = N.arange(ny) - (ny - 1)/2.
        xx, yy = N.meshgrid(x, y)
        xy = xx + 1j * yy
        if step != 1:
            xy *= step
        if rotation:
            xy *= N.exp(1j*rotation)
        if offset:
            xy += offset

        return cls(xy)

    @property
    def nx(self):

        return self.xy.shape[1]

    @property
    def ny(self):

        return self.xy.shape[0]

    @property
    def x(self):
        """Flattened x-coordinate array."""

        return self.xy.real.ravel()

    @property
    def y(self):
        """Flattened y-coordinate array."""

        return self.xy.imag.ravel()

    @property
    def signature(self):
        """
        Return a 4-character signature string such as 'x+y-'.

        Standard order -- *x* increasing along 1st axis and *y* increasing
        along 0th axis -- corresponds to 'x+y+'.  'y-x+' corresponds to
        a transposed coordinate array, with *y* decreasing along the 1st
        axis, and *x* increasing along the 0th axis.

        Noting '+.' a mean finite difference where the real part is
        positive and larger (in absolute value) than the imaginary
        part, one has the following correspondance::

          fd[0] |  +.    -.    .+    .-
          fd[1] |
          ------+-----------------------
            +.  |             y+x+  y-x+
            -.  |             y+x-  y-x-
            .+  | x+y+  x-y+
            .-  | x+y-  x-y-
        """

        # d[0] is the mean finite difference along the columns
        # (axis=1), d[1] is the mean increment along the rows (axis=0)
        fd = [ N.diff(self.xy.mean(axis=axis)).mean() for axis in (0, 1) ]
        xgrad = [ abs(dd.real) > abs(dd.imag) for dd in fd ]
        if xgrad[0] and not xgrad[1]:    # ±. & .±
            axes = 'xy'
        elif not xgrad[0] and xgrad[1]:  # .± & ±.
            axes = 'yx'
        else:
            raise ValueError("Cannot identify grid signature (axes).")

        def sign_major(z):
            """Sign of major component of z."""

            if abs(z.real) > abs(z.imag):  # ±.
                return '+' if z.real > 0 else '-' if z.real < 0 else '0'
            else:                          # .±
                return '+' if z.imag > 0 else '-' if z.imag < 0 else '0'

        signs = [ sign_major(dd) for dd in fd ]  # [±, ±]
        if '0' in signs:
            raise ValueError("Cannot identify grid signature (sort).")

        return ''.join( a+s for a, s in zip(axes, signs) )  # '[xy]±[yx]±'

    def reorder(self, signature):
        """Manipulate self.xy to match signature."""

        assert signature in ('x+y+', 'x+y-', 'x-y+', 'x-y-',
                             'y+x+', 'y+x-', 'y-x+', 'y-x-'), \
            f"Invalid signature {signature!r}."

        try:
            sig_self = self.signature
        except ValueError as error:
            raise ValueError("Cannot reorder unsigned grid.") from error

        # First, is it the same ordering?
        if sig_self[0] != signature[0]:  # No: transpose self.xy
            self.xy = N.transpose(self.xy)
            sig_self = sig_self[2:] + sig_self[:2]  # Update the signature

        # Order along axis 0
        if sig_self[1] != signature[1]:
            self.xy = N.fliplr(self.xy)  # No need to update the signature

        # Order along axis 1
        if sig_self[3] != signature[3]:
            self.xy = N.flipud(self.xy)  # No need to update the signature

        return self.xy

    def plot(self, ax=None, label=None, color='k'):
        """Plot structured grid."""

        import matplotlib.pyplot as P

        if ax is None:
            fig = P.figure()
            ax = fig.add_subplot(xlabel="x", ylabel="y", title=str(self))
            ax.set_aspect('equal', adjustable='datalim')

        for xx, yy in zip(self.xy.real, self.xy.imag):  # Loop over rows
            ax.plot(xx, yy,
                    color=color, marker='.', label=label)
            if label:                 # Only one label
                label = None
        for xx, yy in zip(self.xy.real.T, self.xy.imag.T):  # Over columns
            ax.plot(xx, yy,
                    color=color, marker='.', label='_')

        return ax

    def estimate_parameters(self, rmin=0.1, frac=0.1, rescale=False, fig=None):
        """
        Estimate regular grid parameters from triangulation analysis.

        :param float rmin: minimal circle ratio (see
            :func:`matplotlib.tri.TriAnalyzer.get_flat_tri_mask`)
        :param float frac: fraction of edges used for distortion parameters
        :param bool rescale: rescale long edges
            (not trustworthy for significantly distorted grid)
        :param matplotlib.pyplot.Figure fig: produde a control plot if not None
        :return: (step, rotation [rad],
            complex offset, complex center of distortion)
        """

        from matplotlib.tri import (Triangulation, TriAnalyzer,
                                    LinearTriInterpolator)

        # Triangulation
        tri = Triangulation(self.x, self.y)
        # Optimization: mask border triangles which are too flat
        mask = TriAnalyzer(tri).get_flat_tri_mask(min_circle_ratio=rmin)
        tri.set_mask(mask)

        # Edge lengths
        xy = N.vstack((self.x, self.y)).T  # npts × [x, y]
        edges = xy[tri.edges]              # nedges × [start, end] × [x, y]
        dxy = edges[:, 1] - edges[:, 0]    # nedges × [x, y]
        lengths = N.sum(dxy**2, axis=1)**0.5  # nedges

        # Diagonals (i.e. long edges)
        cut = N.percentile(lengths, 66)  # 2/3 are short edges, 1/3 are long
        medinf = N.median(lengths[lengths <= cut])  # Median of short edges
        medsup = N.median(lengths[lengths >= cut])  # Median of long edges
        med = (medinf + medsup) / 2                 # Separator short/long
        diags = lengths > med

        # Rescale diagonal lengths
        if rescale:
            lengths[diags] /= 2**0.5

        # Angles rewarp in ± pi/4 [rad]
        angles = N.arctan2(dxy[:, 1], dxy[:, 0]) + N.pi  # in [0, 2*pi]
        angles = (4 * angles - N.pi * N.around(4 * angles/N.pi)) / 4

        # Sort lengths by increasing sizes
        iedges = N.argsort(lengths)

        # Use the central fraction of less distorted points (shortest
        # edges) to estimate undistorted grid parameters
        ncen = int(len(edges) * frac)     # Nb of central edges
        shortests = edges[iedges[:ncen]]  # ncen × [start, end] × [x, y]
        step = N.median(lengths[iedges[:ncen]])     # Undistorted grid step
        angle = N.median(angles[iedges[:ncen]])     # Undistorted grid angle
        center = complex(
            *N.mean(shortests, axis=(0, 1)))  # Center of distortion (complex)

        # Reference grid
        ref = StructuredGrid.create(self.nx, self.ny,
                                    step=step, rotation=angle)
        ref.reorder(self.signature)
        offsets = (self.xy - ref.xy).ravel()

        # Interpolate offset at center of distortion
        dx = LinearTriInterpolator(tri, offsets.real)(center.real, center.imag)
        dy = LinearTriInterpolator(tri, offsets.imag)(center.real, center.imag)
        offset = complex(dx, dy)  # Offset (complex)
        ref.xy += offset

        if fig:                 # Create control plot
            import matplotlib.pyplot as P

            if fig is True:     # Create the figure
                fig = P.figure()
            else:               # Use the incoming figure
                assert isinstance(fig, P.Figure)

            axgrid = P.GridSpec(2, 3)
            ax = fig.add_subplot(
                axgrid.new_subplotspec((0, 0),
                                       rowspan=2, colspan=2),
                xlabel="x", ylabel="y",
                title=self.__str__())
            axl = fig.add_subplot(axgrid.new_subplotspec((0, 2)),
                                  xlabel="Side length")
            axa = fig.add_subplot(axgrid.new_subplotspec((1, 2)),
                                  xlabel="Angle [deg]")

            # Input grid
            ax.plot(self.x, self.y, 'o', label='Input')

            # Triangulation
            ax.triplot(tri)
            ax.plot(shortests[..., 0].T, shortests[..., 1].T, 'k-', lw=2)

            # Reference grid
            ax.plot(ref.x, ref.y, '.k', label='Reference')

            # Center of distortion
            ax.plot([center.real], [center.imag],
                    marker='*', ms=10, ls='none', label='Center of distortion')

            ax.legend(loc='best', numpoints=1, fontsize='small',
                      framealpha=0.5, title='')
            ax.set_aspect('equal', adjustable='datalim')

            # Length histogram
            axl.hist([lengths[~diags], lengths[diags]],
                     bins=30, stacked=True, histtype='stepfilled',
                     label=['Short', 'Long/√2'])
            axl.axvline(step, c='k', lw=2)

            # Angle histogram
            axa.hist([N.rad2deg(angles[~diags]),
                      N.rad2deg(angles[diags])],
                     bins=30, stacked=True, histtype='stepfilled',
                     label=['Short', 'Long - 45°'])
            axa.axvline(N.rad2deg(angle), c='k', lw=2)

        return step, angle, offset, center

    def rms(self, xy):
        """
        Compute RMS distance to 2D-array.

        .. Note:: intput grids are supposed to be compatible both in shape and
           signature. RMS cannot be computed or is meaningless otherwise.
        """

        return (N.abs(self.xy - xy)**2).mean()**0.5

    def adjust_distortion(self, other, gdist,
                          scale=False, rotation=False, offset=False,
                          **options):
        """
        Adjust geometric distortion to match other (compatible) grid.

        If *scale* (resp. *rotation* and *offset*), the grid scale
        (resp. rotation and offset) is adjusted simultaneously.

        Other *options* are transmitted to `Minuit` initialization.
        """

        from iminuit import Minuit

        assert self.signature == other.signature, "Incompatible grid signatures."
        assert self.xy.shape == other.xy.shape, "Incompatible grid shapes."

        nkcoeff = len(gdist.Kcoeffs)
        npcoeff = len(gdist.Pcoeffs)

        # Objective function: RMS**2 (see signature below)
        def objfun(*args):

            gdist.x0, gdist.y0 = args[:2]        # Center of distortion
            gdist.Kcoeffs = args[2:2 + nkcoeff]  # K-coeffs
            last = 2 + nkcoeff
            gdist.Pcoeffs = args[last:last + npcoeff]  # P-coeffs
            last += npcoeff
            if scale:
                gdist.scale = args[last]
                last += 1
            if rotation:
                gdist.rotation = args[last]
                last += 1
            if offset:
                gdist.offset = complex(*args[last:last + 2])
                last += 2

            return (N.abs(gdist.forward(self.xy) - other.xy)**2).mean()  # RMS²

        # Objective function parameters
        parameters = ['x0', 'y0'] + \
                     [f'K{i}' for i in range(1, nkcoeff + 1)] + \
                     [f'P{i}' for i in range(1, npcoeff + 1)]

        kwargs = {'name': parameters}  # objfun signature
        errors = {}

        # Estimate position step size from position RMS
        xystep = (N.abs(self.xy)**2).mean()**0.5 * 1e-1
        kwargs.update({'x0': gdist.x0, 'y0': gdist.y0})
        errors.update({'x0': xystep, 'y0': xystep})

        # Arbitrarly set distortion parameter step size to 1e-3
        kwargs.update({key: 0 for key in parameters if key[0] in 'KP'})
        errors.update({f'{key}': 1e-3 for key in parameters if key[0] in 'KP'})

        if scale:
            parameters += ['scale']
            kwargs['scale'] = 1
            errors['scale'] = 1e-2

        if rotation:
            parameters += ['rotation']
            kwargs['rotation'] = 0
            errors['rotation'] = 1e-2

        if offset:
            parameters += ['dx', 'dy']
            # Estimate offset step size from offset RMS
            offstep = self.rms(other.xy)
            kwargs.update({'dx': 0, 'dy': 0})
            errors.update({'dx': offstep, 'dy': offstep})

        # Additional Minuit options
        kwargs.update(**options)

        # Minuit optimizer
        objfun.errordef = Minuit.LEAST_SQUARES
        minuit = Minuit(objfun, **kwargs)
        for k, v in errors.items():
            minuit.errors[k] = v

        # if kwargs.get('print_level', 0) >= 1:
        #     minuit.print_param()

        # Optimization
        minuit.migrad()

        return minuit

    def plot_offsets(self, other, ax=None, units=(1, '')):
        """
        Plot offset between *self* and *other* grid (used as reference).
        """

        import matplotlib.pyplot as P

        uscale, uname = units   # (float, str)

        rms = self.rms(other.xy) / uscale  # Offset RMS [unit]

        if ax is None:
            ustr = f' [{uname}]' if uname else ''
            fig = P.figure()
            ax = fig.add_subplot(1, 1, 1,
                                 xlabel=f"x{ustr}",
                                 ylabel=f"y{ustr}",
                                 title=f"RMS = {rms:.4g} {uname}")
            ax.set_aspect('equal', adjustable='datalim')

        # Offset plot
        dxy = (self.xy - other.xy) / uscale
        q = ax.quiver(other.x / uscale, other.y / uscale,
                      dxy.real, dxy.imag)

        # Quiver key
        scale = f"{rms:.1g}"
        ustr = f' {uname}' if uname else ''
        ax.quiverkey(q, 0.85, 0.95, float(scale), f"{scale}{ustr}",
                     labelpos='W', coordinates='figure')

        return ax


class GeometricDistortion:
    r"""
    Brown-Conrady (achromatic) distortion model.

    .. math::

       x_d &= x_u \times (1 + K_1 r^2 + K_2 r^4 + \ldots) \\
           &+ \left(P_2(r^2 + 2x_u^2) + 2P_1 x_u y_u\right)
              (1 + P_3 r^2 + P_4 r^4 + \ldots) \\
       y_d &= y_u \times (1 + K_1r^2 + K_2r^4 + \ldots) \\
           &+ \left(P_1(r^2 + 2y_u^2) + 2P_2 x_u y_u\right)
              (1 + P_3 r^2 + P_4 r^4 + \ldots)

    where:

    - :math:`z_u = x_u + j y_u` is the undistorted complex position,
    - :math:`z_d = x_d + j y_d` is the distorted complex position,
    - :math:`z_0 = x_0 + j y_0` is the complex center of distortion,
    - :math:`r = |z_u - z_0|` is the undistorted distance to CoD.

    The K-coefficients (resp. P-coefficients) model the *radial*
    (resp. *tangential*) distortion.

    Note there's a possibility to rescale (scale *s* + rotation
    :math:`\alpha` + offset :math:`\delta z`) the input (undistorted)
    positions *before* applying the distortion pattern (see
    :func:`GeometricDistortion.rescale`).

    **Reference:** `Optical distortion
    <https://en.wikipedia.org/wiki/Distortion_%28optics%29>`_

    .. autosummary::

       forward
       backward
       rescale
       unscale
       plot
    """

    def __init__(self, center=0, Kcoeffs=[], Pcoeffs=[],
                 scale=1, rotation=0, offset=0):
        """
        Initialize from center of distortion and K- and P-coefficients.

        :param complex center: complex position of center of distortion [m]
        :param list Kcoeffs: radial distortion coefficients
        :param list Pcoeffs: tangential distortion coefficients
            (empty or length >= 2)
        :param float scale: scaling to be applied to input coordinates
        :param float rotation: rotation to be applied to input coordinates [rad]
        :param complex offset: offset to be applied to input coordinates [m]
        """

        self.center = complex(center)     #: Center of distortion (complex) [m]

        # Radial component
        self._polyk = N.polynomial.Polynomial([1] + list(Kcoeffs))
        # Tangential component
        if Pcoeffs:
            self.p1 = Pcoeffs[0]
            self.p2 = Pcoeffs[1] if len(Pcoeffs) >= 2 else 0
            self._polyp = N.polynomial.Polynomial([1] + list(Pcoeffs)[2:])
        else:
            self._polyp = None  # Will be used as a flag

        self.scale = float(scale)         #: Coordinate scaling
        self.rotation = float(rotation)   #: Coordinate rotation [rad]
        self.offset = complex(offset)     #: Coordinate complex offset [m]

    @classmethod
    def from_kwargs(cls, **kwargs):
        """Initialize from kwargs."""

        center = complex(kwargs.pop('x0', 0), kwargs.pop('y0', 0))
        try:
            kmax = max( int(key[1:]) for key in kwargs if key.startswith('K') )
        except ValueError:
            Kcoeffs = []
        else:
            Kcoeffs = [ kwargs.pop('K'+str(i), 0) for i in range(1, kmax + 1) ]
        try:
            pmax = max( int(key[1:]) for key in kwargs if key.startswith('P') )
        except ValueError:
            Pcoeffs = []
        else:
            Pcoeffs = [ kwargs.pop('P'+str(i), 0) for i in range(1, pmax + 1) ]

        scale = kwargs.pop('scale', 1)
        rotation = kwargs.pop('rotation', 0)
        offset = complex(kwargs.pop('dx', 0), kwargs.pop('dy', 0))

        return cls(center, Kcoeffs=Kcoeffs, Pcoeffs=Pcoeffs,
                   scale=scale, rotation=rotation, offset=offset)

    @property
    def x0(self):
        """x-coordinate of center of distortion."""

        return self.center.real

    @x0.setter
    def x0(self, x0):

        self.center = x0 + 1j * self.center.imag

    @property
    def y0(self):
        """y-coordinate of center of distortion."""

        return self.center.imag

    @y0.setter
    def y0(self, y0):

        self.center = self.center.real + 1j * y0

    @property
    def Kcoeffs(self):
        """Radial coefficients."""

        return self._polyk.coef[1:]       # ndarray

    @Kcoeffs.setter
    def Kcoeffs(self, coeffs):

        self._polyk.coef[1:] = coeffs

    @property
    def Pcoeffs(self):
        """Tangential coefficients."""

        if self._polyp:
            coeffs = [self.p1, self.p2] + self._polyp.coef[1:].tolist()  # list
        else:
            coeffs = []

        return N.array(coeffs)            # ndarray

    @Pcoeffs.setter
    def Pcoeffs(self, coeffs):

        if len(coeffs):         # Should be of length >= 2
            if not len(coeffs) >= 2:
                raise ValueError("coeffs should be empty or of length >= 2.")
            self.p1 = coeffs[0]
            self.p2 = coeffs[1]
            self._polyp.coef[1:] = coeffs[2:]

    def __getattr__(self, attr):
        """Get individual distortion coefficients 'Ki' or 'Pj'."""

        # Kcoeffs and Pcoeffs are handled by the properties defined above
        if attr.startswith('K') and attr != 'Kcoeffs':  # Ki
            # Could raise ValueError or KeyError
            return self._polyk.coef[int(attr[1:])]
        if attr == 'P1':                          # P1
            return self.p1
        elif attr == 'P2':                          # P2
            return self.p2
        elif attr.startswith('P') and attr != 'Pcoeffs':  # Pi
            # Could raise ValueError or KeyError
            return self._polyp.coef[int(attr[1:]) - 2]
        else:
            raise AttributeError()
            # return super().__getattr__(attr)

    def __setattr__(self, attr, value):
        """Set individual distortion coefficients 'Ki' or 'Pj'."""

        # Kcoeffs and Pcoeffs are handled by the properties defined above
        if attr.startswith('K') and attr != 'Kcoeffs':  # Ki
            # Could raise ValueError or KeyError
            self._polyk.coef[int(attr[1:])] = value
        if attr == 'P1':                           # P1
            self.p1 = value
        elif attr == 'P2':                           # P2
            self.p2 = value
        elif attr.startswith('P') and attr != 'Pcoeffs':  # Pi, i > 2
            # Could raise ValueError or KeyError
            self._polyp.coef[int(attr[1:]) - 2] = value
        else:
            super().__setattr__(attr, value)

    def __bool__(self):
        """
        The distortion is non-null iff some coeffs were defined (even if
        null).
        """

        return len(self.Kcoeffs) or len(self.Pcoeffs)

    def __str__(self):

        if self.__bool__():
            s = (f"Geometric distortion: center=({self.x0:+g}, {self.y0:+g}), "
                 f"K-coeffs={self.Kcoeffs}, P-coeffs={self.Pcoeffs}")
        else:
            s = "Null geometric distortion"

        return s

    def rescale(self, xy):
        r"""
        Rescale input complex positions.

        .. math::

           z' = z \times s e^{j\alpha} + \delta z
        """

        return xy * self.scale * N.exp(1j*self.rotation) + self.offset

    def unscale(self, xy):
        r"""
        Unscale output complex positions.

        .. math::

           z = (z' - \delta z) / (s e^{j\alpha})
        """

        return (xy - self.offset) / self.scale / N.exp(1j*self.rotation)

    def forward(self, xyu):
        """
        Apply distortion to undistorted (rescaled) positions.
        """

        xyr = self.rescale(xyu) - self.center  # Relative complex positions
        r2 = N.abs(xyr) ** 2              # Undistorted radii squared
        xu, yu = xyr.real, xyr.imag       # Undistorted coordinates

        xd = N.copy(xu)                   # Distorted coordinates
        yd = N.copy(yu)

        if self.Kcoeffs.any():            # Radial distortion
            polyk_r2 = self._polyk(r2)    # Polynomial in r²
            xd *= polyk_r2
            yd *= polyk_r2

        if self.Pcoeffs.any():            # Tangential distortion
            polyp_r2 = self._polyp(r2)    # Polynomial in r²
            two_xuyu = 2 * xu * yu
            xd += (self.p2 * (r2 + 2 * xu ** 2) + self.p1 * two_xuyu) * polyp_r2
            yd += (self.p1 * (r2 + 2 * yu ** 2) + self.p2 * two_xuyu) * polyp_r2

        xyd = (xd + 1j * yd) + self.center

        return xyd                        # Distorted complex positions

    def backward(self, xyd, lcol=0):
        """
        Correct distortion from distorted complex positions.

        Include a (single-wavelength) lateral color term if needed.
        """

        import scipy.optimize as SO

        xyd = N.atleast_1d(xyd)

        def objf(x_y):
            """
            `SO.root` works on real functions only: we convert N complex
            2D-positions to and fro 2N concatenated real vector.
            """

            xu, yu = N.array_split(x_y, 2)   # (2n,) ℝ → 2×(n,) ℝ
            zu = xu + 1j * yu                # Undistorted position
            zd = self.forward(zu)            # Modeled distorted position
            zd += zu * lcol                  # Add lateral color
            off = zd - xyd.ravel()           # Offsets (n,) ℂ
            return N.concatenate((off.real, off.imag))  # (n,) ℂ → (2n,) ℝ

        # Initial guess: start from distorted position
        xd, yd = xyd.real.ravel(), xyd.imag.ravel()     # → 2×(n,) ℝ
        result = SO.root(objf, N.concatenate((xd, yd)))

        if not result.success:
            raise RuntimeError("GeometricDistortion model is not invertible.")

        xu, yu = N.array_split(result.x, 2)  # Adj. undist. pos. → 2×(n,) ℝ
        zu = (xu + 1j * yu).reshape(xyd.shape).squeeze()

        return self.unscale(zu)  # Remove rescaling

    def plot(self, xy=None, ax=None):
        """
        Plot distortions for a 2D-grid of complex positions.
        """

        import matplotlib.pyplot as P

        if xy is None:
            x = N.linspace(-1.5, 1.5, 13)   # Default grid
            xx, yy = N.meshgrid(x, x)
            xy = xx + 1j * yy               # Undistorted positions
        else:
            assert N.ndim(xy) == 2, "Invalid input grid"

        xyd = self.forward(xy)              # Distorted positions

        # Test
        try:
            test = N.allclose(self.backward(xyd), xy)
        except RuntimeError as err:         # Could not invert distortion
            warnings.warn(str(err))
        else:
            if not test:
                warnings.warn("GeometricDistortion inversion is invalid.")

        if ax is None:
            fig = P.figure()
            title = ' '.join((
                f"K-coeffs={self.Kcoeffs} "
                if self.Kcoeffs.any() else '',
                f"P-coeffs={self.Pcoeffs}"
                if self.Pcoeffs.any() else ''))
            ax = fig.add_subplot(1, 1, 1,
                                 title=title,
                                 xlabel="x", ylabel="y")

        def plot_grid(ax, xy, label=None, color='k'):
            for xx, yy in zip(xy.real, xy.imag):
                ax.plot(xx, yy,
                        color=color, marker='.', label=label)
                if label:                 # Only one label
                    label = None
            for xx, yy in zip(xy.T.real, xy.T.imag):
                ax.plot(xx, yy,
                        color=color, marker='.', label='_')

        # Undistorted positions
        plot_grid(ax, xy, label='Undistorted', color='0.8')
        # Distorted grid
        plot_grid(ax, xyd, label='Distorted', color='k')
        # Center of distortion
        ax.plot([self.center.real], [self.center.imag],
                ls='None', marker='*', ms=10, label='Center')
        ax.set_aspect('equal', adjustable='datalim')
        ax.legend(loc='lower left', fontsize='small',
                  framealpha=0.5, title='')

        return ax


class ChromaticDistortion:
    """
    A polynomial description of Transverse Chromatic Distortion.

    The Transverse Chromatic Aberration (so-called *lateral color*)
    occurs when different wavelengths are focused at different positions
    in the focal plane.

    **Reference:** `Chromatic aberration
    <https://en.wikipedia.org/wiki/Chromatic_aberration>`_

    **See also:** Klein, Brauers & Aach, 2010, for a more detailed modeling of
    Transversal Chromatic Aberrations.

    .. autosummary::

       coeffs
       amplitude
    """

    def __init__(self, wref=1e-6, coeffs=[]):
        r"""
        Initialize from reference wavelength and lateral color coefficients.

        .. math::

           dr = \sum_{i=1}^N c_i (\lambda - \lambda_{\mathrm{ref}})^i

        Note that :math:`c_0 = 0`.

        :param float wref:
            reference wavelength :math:`\lambda_{\mathrm{ref}}` [m]
        :param list coeffs:
            lateral color coefficients :math:`[c_{i=1, \ldots N}]`
        """

        self.wref = wref               #: Reference wavelength [m]
        self._poly = N.polynomial.Polynomial([0] + list(coeffs))

    @property
    def coeffs(self):
        r"""Expose non-null coefficients :math:`[c_{i \geq 1}]`."""

        return self._poly.coef[1:]  # ndarray

    @coeffs.setter
    def coeffs(self, coeffs):

        self._poly.coef[1:] = coeffs

    def __getattr__(self, attr):
        """Get individual distortion coefficients 'Ci'."""

        if attr.startswith('C'):                    # Ci
            # Could raise ValueError or KeyError
            return self._poly.coef[int(attr[1:])]
        else:
            return super().__getattr__(attr)

    def __setattr__(self, attr, value):
        """Set individual distortion coefficients 'Ci'."""

        if attr.startswith('C'):                     # Ci
            # Could raise ValueError or KeyError
            self._poly.coef[int(attr[1:])] = value
        else:
            super().__setattr__(attr, value)

    def __bool__(self):

        return bool(self.coeffs.any())

    def __str__(self):

        if self.__bool__():
            scoeffs = ', '.join( f'{coeff:+g}' for coeff in self.coeffs )
            s = (f"Chromatic distortion: lref={self.wref / 1e-6:.2f} µm, "
                 f"coeffs=[{scoeffs}]")
        else:
            s = "Null chromatic distortion"

        return s

    def amplitude(self, wavelengths):
        """
        Compute amplitude of the (radial) chromatic distortion.

        :param numpy.ndarray wavelengths: wavelengths [m]
        :return: lateral color radial amplitude
        """

        return self._poly(wavelengths / self.wref - 1)
